from pioneer.das.api.sensors.sensor  import Sensor
from pioneer.das.api.samples         import PointCloud
from pioneer.das.api.interpolators   import nearest_interpolator

from typing import Any, Dict, Tuple, Union

import numpy as np
import warnings
warnings.simplefilter('once', DeprecationWarning)


class Sensor3D(Sensor):
    
    def __init__(self, name: str, platform: 'Platform', factories:Dict[str, Tuple[Any, Any]] = {}):
        super().__init__(name, platform, factories={**factories, 'pcloud': (PointCloud, nearest_interpolator)})

    def load_intrinsics(self, intrinsics_config: Union[str, dict]):
        '''load data from yml platform '''
        if isinstance(intrinsics_config, str):
            super().load_intrinsics(intrinsics_config)
        orientation = intrinsics_config.get('orientation')
        if orientation: self.orientation = np.array(orientation)


from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.samples import Image, ImageFisheye
from pioneer.das.api.sensors.sensor import Sensor
from pioneer.das.api.interpolators import nearest_interpolator

import numpy as np

class Camera(Sensor):
    """Camera sensor, expects 'img' datasource, intrinsics matrix and distortion coefficients"""

    def __init__(self, name: str, platform: 'Platform'):
        super(Camera, self).__init__(name
                                   , platform
                                   , { 'img': (Image, nearest_interpolator),
                                       'flimg': (ImageFisheye, nearest_interpolator)
                                        })

    @property
    def camera_matrix(self):
        """np.ndarray: the 3x3 intrinsics matrix
        
        See also: https://docs.opencv.org/4.0.0/d9/d0c/group__calib3d.html#ga3207604e4b1a1758aa66acb6ed5aa65d
        """ 
        try:
            matrix = self.intrinsics['matrix']
        except:
            LoggingManager.instance().warning(f'Intrinsic matrix of {self.name} not found. Trying to make a generic one.')
            h = self.yml['configurations']['img']['Width']
            v = self.yml['configurations']['img']['Height']
            h_fov = self.yml['configurations']['img']['h_fov']
            matrix = np.identity(3)
            matrix[0,2] = h/2
            matrix[1,2] = v/2
            matrix[0,0] = matrix[1,1] = h/(2*np.tan(h_fov * np.pi / 360.0))
        return matrix

    @property
    def distortion_coeffs(self):
        """np.ndarray: Nx1 distortion coefficient, refer to opencv documentation 
        
        See also: https://docs.opencv.org/4.0.0/d9/d0c/group__calib3d.html#ga3207604e4b1a1758aa66acb6ed5aa65d
        """
        try:
            k = self.intrinsics['distortion']
        except:
            LoggingManager.instance().warning(f'Distortion coeffients not found for {self.name}.')
            k = np.zeros((5,1))
        return k

    def load_intrinsics(self, intrinsics_config):
        super(Camera, self).load_intrinsics(intrinsics_config)
        if self.intrinsics is not None:
            assert 'matrix' in self.intrinsics
            # previous calibration had a typo in the distortion coeffs
            if (not 'distortion' in self.intrinsics and
                'distorsion' in self.intrinsics):
                self.intrinsics['distortion'] = self.intrinsics['distorsion']
            assert 'distortion' in self.intrinsics



from pioneer.das.api.datasources import DataSource

import numpy as np

class AbstractDatasource(DataSource):
    def __init__(self, sensor:'Sensor', ds_type:str, is_live:bool = False):
        self.sensor = sensor
        self.ds_type = ds_type
        self.config = {}
        self.is_live = is_live


    def __len__(self) -> int:
        return len(self.timestamps)

    def to_positive_index(self, index):
        if index < 0:
            index = index % len(self)
        return index
        
    @property
    def timestamps(self) -> np.ndarray:
        raise RuntimeError("Not implemented")

    @property
    def time_of_issues(self) -> np.ndarray:
        raise RuntimeError("Not implemented")

    @property
    def label(self) -> str:
        return "{}_{}".format(self.sensor.name, self.ds_type)

    def to_float_index(self, timestamp:int) -> float:
        i = np.searchsorted(self.timestamps, timestamp)
        # deal with borders
        if i == 0:
            return 0
        if i == len(self.timestamps):
            return i-1
        ts_from = self.timestamps[i - 1]
        t = float(timestamp - ts_from) / (self.timestamps[i] - ts_from)
        if t == 0:
            return i - 1
        if t == 1:
            return i

        return i - 1 + t


from pioneer.common import mercator_projection
from pioneer.das.api.samples.image import Image

import cv2
import numpy as np

class ImageFisheye(Image):
    '''A derivation the Image sample, to be used with fisheye lenses.'''
    
    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super().__init__(index, datasource, virtual_raw, virtual_ts) 
        if not hasattr(self.datasource.sensor, 'mercator_projection'):
            self.datasource.sensor.mercator_projection = mercator_projection.MercatorProjection(self.camera_matrix, self.distortion_coeffs) 

    def project_pts(self, pts, mask_fov=False, output_mask=False, undistorted=False, margin=0):

        R = T = np.zeros((3, 1))
        
        if undistorted or mask_fov or output_mask:
            A, k = self.und_camera_matrix, np.zeros((5,1))
            und_image_pts = self.datasource.sensor.mercator_projection.project_pts(pts)
            und_image_pts = np.squeeze(und_image_pts)

        if not undistorted:
            A, k = self.camera_matrix, self.distortion_coeffs
            image_pts, _ = cv2.fisheye.projectPoints(pts.reshape((-1,1,3)), R, T, A, k)
            image_pts = np.squeeze(image_pts)
        else:
            image_pts = und_image_pts

        if mask_fov or output_mask:
            mask = self.projection_mask(pts, und_image_pts, margin)
            
        if mask_fov:
            image_pts = image_pts[mask]

        if output_mask:
            return image_pts, mask
        return image_pts

    def get_image(self, undistort: bool = False) -> np.ndarray:
        if undistort:
            return self.datasource.sensor.mercator_projection.undistort(self.get_image(undistort=False))
        return super().get_image()


from typing import Optional
from pioneer.common import platform, linalg
from pioneer.common.logging_manager import LoggingManager

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pioneer.das.api.egomotion import EgomotionProvider

import numpy as np
from numpy import rot90, flipud, fliplr #for use with Sample.LUT
import pprint
import six

def warn_if_less_than_64bit(dtype:np.dtype):
    if np.finfo(dtype).bits < 64:
        LoggingManager.warning(f"{dtype} precision detected, performing computations in that context with less than 64bits precision is ill-advised")


# TODO: Move all 3D stuff to Sensor3D class
# TODO: Move all 2D image stuff to Image class
# TODO: Refactor the transform methods
# TODO: See if we can remove the weird LUT stuff


class Sample:

    LUT = {}

    @staticmethod
    def __goc_lut():
        if not Sample.LUT:
            # Basic transformation on xy-plane
            rotations = {
            # Clockwise rotation sends x to -y and y to x
            'rot90({},k=-1)': np.array([[0, -1,  0],
                 [1,  0,  0],
                 [0,  0,  1]], 'f4'),
            # Counter-clockwise rotation sends x to y and y to -x
            'rot90({},k=1)': np.array([[0,  1,  0],
                 [-1, 0,  0],
                 [0,  0,  1]], 'f4'),
            'rot90({},k=2)': np.array([[-1, 0,  0],
                 [0, -1,  0],
                 [0,  0,  1]], 'f4'),
            '{}': np.eye(3, dtype='f4')}

            reflexions = {
            'fliplr({})': np.array([[-1, 0,  0],
                 [0,  1,  0],
                 [0,  0,  1]], 'f4'),
            'flipud({})': np.array([[1,  0,  0],
                 [0, -1,  0],
                 [0,  0,  1]], 'f4'),
            '{}': np.eye(3, dtype='f4')}

            # Other orientations than are not defined on xy plane:
            special_transformations = {
                np.array([[1,  0,  0],
                 [0, 1,  0],
                 [0,  0,  -1]], 'f4').tobytes(): 'flipud({})',
                np.array([[0, 0, 1],
                 [-1, 0, 0],
                 [0, 1, 0]], 'f4').tobytes(): '{}',
                np.array([[0, 0, 1],
                 [-1, 0, 0],
                 [0, -1, 0]], 'f4').tobytes(): 'fliplr({})',

            }

            for rot_op, rot in rotations.items():
                for refl_op, refl in reflexions.items():
                    comb = np.matmul(rot, refl)
                    inv_com = np.matmul(refl, rot)
                    Sample.LUT[comb.tobytes()] = rot_op.format(refl_op)
                    Sample.LUT[inv_com.tobytes()] = refl_op.format(rot_op)

            Sample.LUT.update(special_transformations)

        return Sample.LUT

    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        '''
            virtual_raw and virtual_ts are meant for samples which dont exist in
            a physical datasource. For example interpolated samples, or samples
            derived from other samples.
        '''
        self.index = index
        self.datasource = datasource
        self.virtual_raw = virtual_raw
        self.virtual_ts = virtual_ts
        self._raw = None

    @property
    def orientation(self):
        return self.datasource.sensor.orientation

    @property
    def intrinsics(self):
        return self.datasource.sensor.intrinsics

    @property
    def extrinsics(self):
        return self.datasource.sensor.extrinsics

    @property
    def raw(self):
        if self.virtual_raw is None:
            if self._raw is None:
                self._raw = self.datasource.ds[self.index]
            return self._raw
        return self.virtual_raw

    @property
    def timestamp(self):
        if self.virtual_ts is None:
            return self.datasource.timestamps[self.index]
        return self.virtual_ts

    @property
    def time_of_issues(self):
        if self.virtual_ts is None:
            return self.datasource.time_of_issues[self.index]
        return self.virtual_ts

    @property
    def label(self):
        return self.datasource.label

    @property
    def sensor_type(self):
        return platform.parse_datasource_name(self.datasource.label)[0]

    def compute_transform(self, referential_or_ds:str=None, ignore_orientation:bool=False, reference_ts:int = -1, dtype = np.float64) -> np.ndarray:
        """Compute the transform from this Sample sensor to another sensor
        referential.

        By default the transform between the two sensors reference frames include
        the contribution from the orientation member. If ones wishes to ignore this
        contribution it can set the ignore_orientation parameter to True.

        Args:
            referential_or_ds: The other sensor name or full datasource name, 
                                e.g. flir_tfc or flir_tfc_img (default: {None})
            ignore_orientation: Ignore the source sensor orientation (default: {False})
            reference_ts:   (only used if referential_or_ds == 'world'), the timestamp at which we want to 'jump' from 
                            EgomotionProvider's referential (e.g. the IMU) to the to 'world' referential. Also note that 
                            the 'world' referential is actually EgomotionProdiver's referential at some reference time 't' 
                            (refer to your actual EgomotionProvider's configuration, by default 't' is EgomotionProvider's
                            initial timestamp). 'world' referential is thus commonly refered to as the 'EgoZero' referential

            dtype: the output numpy data type

        Returns:
            The [4, 4] transformation matrix
        """

        # TODO: This needs refactoring...

        tf_TargetRef_from_Local = np.eye(4, dtype = dtype)
       
        try:
            if isinstance(referential_or_ds, six.string_types) and self.datasource.sensor.name != referential_or_ds:
                
                referential = platform.referential_name(referential_or_ds)

                if referential == 'world':

                    warn_if_less_than_64bit(dtype)

                    if reference_ts < 0:
                        reference_ts = self.timestamp

                    provider:'EgomotionProvider' = self.datasource.sensor.platform.egomotion_provider
                    try:
                        tf_Ego_from_Local = self.datasource.sensor.map_to(provider.referential_name)
                    except Exception:
                        tf_Ego_from_Local = tf_TargetRef_from_Local
                    tf_TargetRef_from_Local = provider.get_first_inverse_transform() @ provider.get_transform(reference_ts) @ tf_Ego_from_Local

                else:
                    tf_TargetRef_from_Local = np.copy(self.datasource.sensor.map_to(referential))
        except:
            LoggingManager.instance().warning(f'Transformation matrix between {self.datasource.sensor.name} and {referential_or_ds} could not be computed. Using identity')
                    
        if self.orientation is not None and not ignore_orientation:
            tf_TargetRef_from_Local[:3, :3] = tf_TargetRef_from_Local[:3, :3] @ self.orientation
        
        return tf_TargetRef_from_Local

    def transform(self, pts:np.ndarray, referential_or_ds:str, ignore_orientation:bool=False, reference_ts:int = -1, reverse:bool=False, dtype = np.float64) -> np.ndarray:
        """Transform 3D points from this Sample sensor to another sensor
        referential.

        Arguments:
            pts: The [N, 3] points to be transformed
            referential_or_ds: The target sensor referential or full datasource name
            ignore_orientation: Ignore the source sensor orientation (default: {False})
            reference_ts: refer to compute_transform()'s doc (only used if referential_or_ds == 'world')
            reverse: apply the reverse transformation
            dtype: the output numpy data type

        Returns:
            The transformed points
        """
        r = self.compute_transform(referential_or_ds, ignore_orientation=ignore_orientation, reference_ts=reference_ts, dtype=dtype)

        if reverse:
            r = linalg.tf_inv(r)

        if r is not None:
            return Sample.transform_pts(r, pts)

        return pts

    def undistort_points(self, points:np.ndarray, timestamps:np.ndarray, reference_ts:Optional[int] = -1, to_world:bool = False, dtype = np.float64):
        """Transform 3D points that have not been sampled simultaneously to their 'correct' place
        referential.

        Args:
            points: array [N, 3] of points to be transformed
            timestamps: the N timestamps (common for all point batches)
            to_world:   If 'True', leave undistorted points in 'world' referential, otherwise
                        project them back to local referential
            reference_ts:   only used if to_world == False, let the use chose at what time
                            undistorted points are projected back to the local referential
                            (useful to compare points from different sensors in a common local referential)
            dtype: the output numpy data type

        Returns:
            The transformed points
        """

        warn_if_less_than_64bit(dtype)

        provider:'EgomotionProvider' = self.datasource.sensor.platform.egomotion_provider        
        to_egomotion_provider_referential = self.compute_transform(provider.referential_name, False, dtype = dtype)
        subsampling = 100
        trajectory = provider.get_trajectory(timestamps, subsampling=subsampling)

        point_ego_referential = linalg.map_points(to_egomotion_provider_referential, points)

        corrected = np.empty_like(point_ego_referential)
        n_points = point_ego_referential.shape[0]
        for i in range(trajectory.shape[0]):
            s, e = i * subsampling, min((i+1) * subsampling, n_points)
            corrected[s:e] = linalg.map_points(trajectory[i], point_ego_referential[s:e])

        if to_world:
            points[:] = corrected
        else:
            if reference_ts < 0: reference_ts = self.timestamp
            to_world = provider.get_inverse_transform(reference_ts)
            tf_Local_from_EgoZero = linalg.tf_inv(to_egomotion_provider_referential) @ to_world @ provider.get_first_transform()
            points[:] = linalg.map_points(tf_Local_from_EgoZero, corrected)


    def _get_orientation_lut(self):
        if self.orientation is not None:
            orientation_f4 = self.orientation.astype('f4')
            lut = Sample.__goc_lut()
            if orientation_f4.tobytes() in lut:
                return lut[orientation_f4.tobytes()]

            LoggingManager.instance().warning('orientation {} could not be mapped to any image transform'.format(self.orientation))
        return None

    def transform_image(self, image):
        # Transform the bottom left origin of LCAx to top-left camera origin
        image = np.flipud(image)

        operation = self._get_orientation_lut()
        if operation is not None:
            return np.copy(eval(operation.format('image')))

        return np.copy(image)

    @staticmethod
    def transform_pts(matrix4x4, ptsNx3):
        return linalg.map_points(matrix4x4, ptsNx3)

    def pretty_print(self):
        return pprint.pformat(self.raw)



from pioneer.common import platform
from pioneer.das.api import categories
from pioneer.das.api.samples.sample import Sample

import cv2
import numpy as np

class Poly2d(Sample):

    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super(Poly2d, self).__init__(index, datasource, virtual_raw, virtual_ts)

    def colored_image(self, resolution:tuple=None):
        polygons = self.raw
        _,_,ds_type = platform.parse_datasource_name(self.datasource.label)
        poly_source = categories.get_source(ds_type)
        image = np.zeros((polygons['resolution'][0],polygons['resolution'][1],3), dtype=np.uint8)
        
        for poly in polygons['data']:
            name, color = categories.get_name_color(poly_source,poly['classes'])
            color = np.array(color)/255
            cv2.fillPoly(image, [poly['polygon']], color)

        return self.resize_mask(image, resolution) if resolution is not None else image

    def mask_category(self, category:str, resolution:tuple=None, confidence_threshold:float=0.5):
        polygons = self.raw
        _,_,ds_type = platform.parse_datasource_name(self.datasource.label)
        poly_source = categories.get_source(ds_type)
        mask = np.zeros((polygons['resolution'][0],polygons['resolution'][1]), dtype=np.uint8)
        
        for i, poly in enumerate(polygons['data']):

            if 'confidence' in polygons:
                if polygons['confidence'][i] < confidence_threshold:
                    continue

            name, color = categories.get_name_color(poly_source,poly['classes'])
            if name == category:
                cv2.fillPoly(mask, [poly['polygon']], 1)

        return self.resize_mask(mask, resolution) if resolution is not None else mask

    def areas(self):
        areas = np.empty(0)
        for poly in self.raw['data']:
            areas = np.append(areas, cv2.contourArea(poly['polygon']))
        return areas

    @staticmethod
    def resize_mask(mask, resolution):
        return cv2.resize(mask, (resolution[1],resolution[0]), interpolation = cv2.INTER_NEAREST)



from pioneer.common                  import platform, clouds
from pioneer.das.api                 import categories
from pioneer.das.api.samples.sample  import Sample

import numpy as np

class Seg3d(Sample):

    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super(Seg3d, self).__init__(index, datasource, virtual_raw, virtual_ts)
            
    def colors(self, mode=None):

        _,_,ds_type = platform.parse_datasource_name(self.datasource.label)
        seg_source = categories.get_source(ds_type)
        classes = self.raw['data']['classes']

        colors = np.zeros((len(classes),4))

        for c in np.unique(classes):
            name, color = categories.get_name_color(seg_source, c)
            color = np.array(color)/255. # color channels between 0 and 1
            color = np.append(color, 1) # alpha (opacity) = 1
            ind = np.where(classes == c)[0]
            colors[ind] = color

        if mode=='quad_cloud':
            colors = clouds.quad_stack(colors)

        return colors

    def mask_category(self, category):
        classes = self.raw['data']['classes']

        if type(category) is int:
            return classes == category

        elif type(category) is str:
            _,_,ds_type = platform.parse_datasource_name(self.datasource.label)
            seg_source = categories.get_source(ds_type)
            try:
                category_number = categories.get_category_number(seg_source, category)
            except:
                category_number = -1
            return classes == category_number
        
        else:
            raise ValueError('The category must be either an integer or a string.')
            
          


from pioneer.das.api.samples.sample import Sample


class Lane(Sample):
    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super(Lane, self).__init__(index, datasource, virtual_raw, virtual_ts)  


import glob
import multiprocessing
import os
import shutil
import tqdm
import zipfile



def __extract_or_copy_file(args):
    filepath, output = args
    try:
        if filepath.endswith('.zip'):
            filename = filepath.split('/')[-1].split('.')[0]
            archive = zipfile.ZipFile(filepath, 'r')
            archive.extractall(f'{output}/{filename}/.')
        elif os.path.isdir(filepath):
            name = filepath.split('/')[-1]
            shutil.copytree(filepath, f'{output}/{name}')
        else:
            name = filepath.split('/')[-1]
            shutil.copyfile(filepath, f'{output}/{name}')
    except:
        print(f'Failed to extract or copy: {filepath}')

def extract_zipfiles(dataset:str, output:str, files:list='all'):
    """Make a copy of a dataset with unzipped files
    
        Args:
            dataset (str): Path to the original dataset
            output (str): Path to the copy to be created. Must be non-existant for safety reasons.
            files (list): list of zipfile names to include. Default is 'all'.
    """

    if not os.path.exists(output):
        os.mkdir(output)
    else:
        print(f"Cannot extract to {output}, because it is an already existing directory.")
        return

    all_files = glob.glob(f'{dataset}/*')
    
    if isinstance(files, list):
        all_files = [f for f in all_files if f.split('/')[-1] in files]

    all_files = [[f, output] for f in all_files]

    pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())
    for _ in tqdm.tqdm(pool.imap_unordered(__extract_or_copy_file, all_files), total = len(all_files)):
        pass


from pioneer.common import platform as platform_utils
from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.datasources import AbstractDatasource
from pioneer.das.api.datasources import virtual_datasources
from pioneer.das.api.egomotion import EgomotionProvider
from pioneer.das.api.samples.sample import Sample
from pioneer.das.api.sensors import Sensor
from pioneer.das.api.sensors import SENSOR_FACTORY
from pioneer.das.api.sources import DirSource, ZipFileSource

from typing import Iterator, Union, Optional, List, Dict, Mapping, Tuple, Any

import ast
import glob
import numbers
import numpy as np
import os
import pandas as pd
import six
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import sys
import tqdm
import yaml
import zipfile

def parse_yaml_string(ys):
    fd = StringIO(ys)
    dct = yaml.safe_load(fd)
    return dct
    
try:
    import numba
    HAVE_NUMBA = True
except ImportError:
    LoggingManager.instance().warning('Numba is not installed, you may have MemoryErrors when '
                  'synchronizing data sources.')
    HAVE_NUMBA = False

def closest_timestamps_np(ref_ts:np.ndarray, target_ts:np.ndarray, tol:Union[float, int]):
    """Finds indices of timestamps pairs where a value in 'target_ts' is whithin 'tol' of a value 'ref_ts'

        Args:
            ref_ts: the timestamps of the sensor we want to synchronize with
            target_ts: the timestamps of the sensor we hope to find matches with 'ref_ts' whithin 'tol'
        Returns:
            The indices of the matches
    """
    try:
        diff = np.abs(ref_ts[:, None].astype('i8') -
                        target_ts[None, :].astype('i8'))
    except MemoryError:
        print("You don't have enough memory. "
              "Try installing numba to resolve this issue.\n"
              "pip3 install numba or conda install numba",
              file=sys.sterr
              )
        raise

    # find the closest leddar timestamp for each camera timestamp
    idx = np.argmin(diff, axis=1)

    # what if many images match the same leddar data package equally?
    min_diff = diff[np.arange(diff.shape[0]), idx]
    too_large = min_diff > tol
    idx[too_large] = -1
    return idx

closest_timestamps = closest_timestamps_np

if HAVE_NUMBA:
    @numba.njit(cache=True)
    def closest_timestamps_numba(ref_ts:np.ndarray, target_ts:np.ndarray, tol:Union[float, int]):
        """ Numba-optimized version of closest_timestamps_np()"""
        n_ref = ref_ts.shape[0]
        ref_ts = ref_ts.astype(np.int64)
        target_ts = target_ts.astype(np.int64)
        indices = np.full(n_ref, -1)

        for i_ref in range(n_ref):
            diff = np.abs(ref_ts[i_ref] - target_ts)
            min_idx = np.argmin(diff)
            if diff[min_idx] > tol:
                continue
            indices[i_ref] = min_idx

        return indices
        
    closest_timestamps = closest_timestamps_numba

class Platform:
    """A Platform is what encapsulate an instance (configuration) of some data acquisition platform. 
       It contains one or more sensors, each containing one or more datasources. A live platform interfaces
       live sensors, while an offline platform can be used to extract data from a recording. 
    """

    def __init__(self, dataset:Optional[str]=None, configuration:Optional[str]=None, include:Optional[list]=None, ignore:Optional[list]=[],
     progress_bar:bool=True, default_cache_size:int=100, activate_logger:bool=False):
        """Constructor

           Args:
            dataset:    If None, this platform will be considered a live sensor platform. 
                        Otherwise, 'dataset' contains the path to an offline recording, where 
                        one expects to find one or many '.zip' filesources named like their 
                        corresponding datasource. If argument 'configuration' is None, then one
                        expect to find a file named 'platform.yml' which declares which sensor this
                        platform must consider, and thus which '.zip' file to use. 'zip' files are 
                        expected to contain one file name according to das.api.filesource.TIMESTAMPS_CSV_PATTERN, 
                        zero or one file name according to das.api.filesource.CONFIG_YML_PATTERN, and 
                        other files corresponding to actual sensor data (e.g. '.pkl' files, '.png' files, etc)
            configuration:  If None, see 'dataset' documentation. Otherwise, 'configuration' is expected to contain
                            a path to a '.yml' file describing a das.api acquisition platform configuration, or its content
                            directly, which allows the user to instanciate a platform without creating a file. 
                            If 'dataset' is not None, 'configuration' can be a path relative to 'dataset'
            include:    A list of strings (e.g. ['lca2','flir_tfc']) to be included exclusively in the platform's keys
            ignore:     A list of strings (e.g. ['lca2','flir_tfc']) to be ignored and not included in the platform's keys
            progress_bar:   If True, show the progress bars when initializing or synchronizing.
            default_cache_size: the default value for (offline) datasource cache size
        """

        LoggingManager.instance().set_verbosity(activate_logger)

        self.dataset = dataset
        self.progress_bar = progress_bar
        self.default_cache_size = default_cache_size
        
        self.configuration = configuration
        if self.configuration is None:
            self.configuration = glob.glob(os.path.join(dataset, 'platform.yml'))[0]

        is_config_a_file = os.path.exists(self.configuration) or os.path.isabs(self.configuration)

        if not is_config_a_file and self.dataset is not None:
            candidate = os.path.join(self.dataset, self.configuration)
            is_config_a_file = os.path.exists(candidate) or os.path.isabs(candidate)
            if is_config_a_file:
                self.configuration = candidate

        if is_config_a_file:
            with open(self.configuration, 'r') as f:
                self.yml = yaml.safe_load(f)
        else:
            self.yml = parse_yaml_string(self.configuration)

        self.include = include
        self.ignore = ignore

        # If there is a way to use wildcards here instead, it would be better
        include = self.yml.keys() if include is None else include
        to_keep, to_pop = [], []
        for additionnal_key in ['virtual_datasources', 'synchronization']:
            if additionnal_key in self.yml:
                to_keep.append(additionnal_key)
        for sensor in self.yml:
            for inc in include:
                if inc in sensor:
                    to_keep.append(sensor)
            for ign in ignore:
                if ign in sensor:
                    to_pop.append(sensor)
        self.yml = {sensor:self.yml[sensor] for sensor in to_keep}
        for sensor in set(to_pop):
            try:
                self.yml.pop(sensor)
            except:
                pass

        self.metadata_path = f'{dataset}/metadata.csv'
        if os.path.exists(self.metadata_path):
            self.metadata = pd.read_csv(self.metadata_path, index_col=0, converters={'keywords':ast.literal_eval})
            self.metadata = self.metadata.where(pd.notnull(self.metadata), None)
        else:
            self.metadata = None
        if os.path.exists(f"{dataset}/manual_c_metadata.zip"):
            self.yml["manual_c"] = ""

        self._sensors = Sensors(self, self.yml)

        if 'virtual_datasources' in self.yml:
            self.add_virtual_datasources(self.yml['virtual_datasources'])

        if os.path.isfile(os.path.join(dataset, 'categories.yml')):
            from pioneer.das.api import categories
            categories.append_categories(os.path.join(dataset, 'categories.yml'))

    def to_nas_path(self, path:str) -> str:
        """Convert absolute yaml paths to path relative to os.environ['nas'] """
        nas_base = os.environ.get('nas', '')
        return nas_base + path # do not use os.path.join() here, as nas defaults to /nas, which is absolute

    def from_relative_path(self, relative_path:str) -> str:
        """Converts a path relative to dataset folder to an absolute path """

        if self.dataset is None:
            raise RuntimeError("dataset path is None")

        return os.path.join(self.dataset, relative_path)

    def try_absolute_or_relative(self, folder:str) -> str:
        """ tries for absolute or relative path """

        folder_path = self.to_nas_path(folder)

        if not os.path.exists(folder_path):
            folder_path = self.from_relative_path(folder)

        if not os.path.exists(folder_path):
            raise RuntimeError(f"Invalid extrinsics path: {folder}")

        return folder_path

    @property
    def sensors(self) -> 'Sensors':
        """Returns this platform's Sensors instance"""
        return self._sensors

    @property
    def orientation(self) -> Dict[str, Optional[np.ndarray]]:
        """Returns a dict with orientation matrix for each platform's sensor"""
        return {sensor.name: sensor.orientation for sensor in self._sensors.values()}

    @property
    def intrinsics(self) -> Dict[str, Any]:
        """Returns a dict with intrinsics information for each platform's sensor"""
        return {sensor.name: sensor.intrinsics for sensor in self._sensors.values()}

    @property
    def extrinsics(self) -> Dict[str, Mapping[str, np.ndarray]]:
        """Returns a dict with extrinsics information for each platform's sensor"""
        return {sensor.name: sensor.extrinsics for sensor in self._sensors.values()}

    def expand_wildcards(self, labels:List[str]) -> List[str]:
        """See also: platform.expand_widlcards()
        """ 
        return platform_utils.expand_wildcards(labels, self.datasource_names())

    def synchronized(self, sync_labels:List[str]=[], interp_labels:List[str]=[], tolerance_us:Union[float, int]=None, fifo:int=-1):
        """Creates a Synchronized instance with self as platform
           
           See Also: Synchronized.__init__()
        """
        if 'synchronization' in self.yml:
            sync_labels = self.yml['synchronization']['sync_labels'] if len(sync_labels)==0 else sync_labels
            interp_labels = self.yml['synchronization']['interp_labels'] if len(interp_labels)==0 else interp_labels
            tolerance_us = self.yml['synchronization']['tolerance_us'] if tolerance_us is None else tolerance_us
        tolerance_us = 1e3 if tolerance_us is None else tolerance_us
        return Synchronized(self, self.expand_wildcards(sync_labels)
                            , self.expand_wildcards(interp_labels)
                            , tolerance_us, fifo)

    def is_live(self) -> bool:
        """Returns wether this platform contains live sensors or offline recordings"""
        return self.dataset is None

    def start(self):
        """**Live platform only** starts the (live) sensors"""
        self._sensors.start()

    def stop(self):
        """**Live platform only** stops the (live) sensors"""
        self._sensors.stop()
    
    def record(self):
        """**Live platform only** toggle recording for all (live) sensors"""
        for ds_name in self.datasource_names():
            try:
                self[ds_name].ds.is_recording = not self[ds_name].ds.is_recording
            except: pass
    
    @property
    def egomotion_provider(self) -> EgomotionProvider:
        return self._sensors._egomotion_provider

    def datasource_names(self) -> List[str]:
        """Returns the list of all datasource names, e.g. ['pixell_tfc_ech', 'pixell_tfc_sta', ...]"""
        ds_names = []
        for s in self._sensors.values():
            ds_names.extend(s.datasource_names())
        return ds_names

    def add_egomotion_provider(self, egomotion_provider:EgomotionProvider):
        self[egomotion_provider._referential_name].egomotion_provider = egomotion_provider
        self._sensors._egomotion_provider = egomotion_provider
        egomotion_provider.sensor = self[egomotion_provider._referential_name]

    def add_virtual_datasources(self, virtual_datasources_config:Dict[str, Dict]):
        """Add virtual datasources based upon the provided configuration dictionnary
        
            Args:
                -virtual_datasources_config (dict): Each key must correspond to a key available in
                    VIRTUAL_DATASOURCE_FACTORY (see api/datasources/virtual_datasources/__init__.py).
                    Under each key, a dictionnary must be provided with the parameters necessary
                    to create the corresponding virtual datasource. These parameters always contain at 
                    least the 'reference_sensor' and 'dependencies' entries. Other entries depend upon
                    the specific virtual datasource type.

                    For exemple:
                        virtual_datasources_config = {
                            'echoes_from_traces': {
                                'reference_sensor': 'pixell_bfc',
                                'dependencies': ['pixell_bfc_ftrr'],
                                'nb_detections_max': 3,
                            }
                        }

            Note: To create multiple virtual datasources of the same type, you can do something like:
                virtual_datasources_config = {
                    'echoes_from_traces_bfc': {'reference_sensor': 'pixell_bfc', 'dependencies': ['pixell_bfc_ftrr']}
                    'echoes_from_traces_tfc': {'reference_sensor': 'pixell_tfc', 'dependencies': ['pixell_tfc_ftrr']}
                }
        """

        for virtual_ds_name in virtual_datasources_config:
            args = virtual_datasources_config[virtual_ds_name]

            # If multiple instances of the same VirtualDatasource class, their keys has to be different
            # in the config file. So add a unique id such as: virtual_ds_name -> virtual_ds_name_id
            if hasattr(virtual_datasources, virtual_ds_name[:virtual_ds_name.rfind('_')]):
                virtual_ds_name = virtual_ds_name[:virtual_ds_name.rfind('_')]

            if hasattr(virtual_datasources, virtual_ds_name):
                try:
                    virtual_datasource = virtual_datasources.VIRTUAL_DATASOURCE_FACTORY[virtual_ds_name](**args)
                    self[virtual_datasource.reference_sensor].add_datasource(virtual_datasource, virtual_datasource.ds_type)
                except:
                    LoggingManager.instance().warning(f"The virtual datasource {virtual_ds_name} could not be added.")
            else:
                LoggingManager.instance().warning(f"The virtual datasource {virtual_ds_name} does not exist.")
        
    def __getitem__(self, label:str) -> Union[Sensor, AbstractDatasource]:
        """Implement operator '[]'

           Args:
            label: can be a sensor name (e.g. 'pixell_tfc') or a datasource name (e.g. 'pixell_tfc_ech')
           Returns:
            the Sensor or AbstractDatasource instance
           Raises:
            IndexError: label must contain exactly 2 or 3 '_' 
        """
        parts = label.split('_')
        if len(parts) == 2:
            return self.sensors[label]
        if len(parts) == 3:
            return self.sensors[parts[0] + '_' + parts[1]][parts[2]]
        else:
            raise IndexError('Invalid label: {}'.format(label))


class Synchronized:
    """Synchronized view over a Platform"""

    def __init__(self, platform:Platform, sync_labels:List[str], interp_labels:List[str], tolerance_us:Union[float, int], fifo:int):
        """Constructor

            Args:
                platform:       The platform that contains datasources to synchronize
                sync_labels:    A list of expanded datasource names to harvest tuples of samples synchronized 
                                whithin 'tolerance_us' microseconds. **Important** The first label will serve as the 'reference 
                                datasourse'
                interp_labels:  A list of expanded datasource names from which to obtain interpolated samples 
                                to be added to the 'synchronized tuples'. 
                                The corresponding datasource must have an interpolator defined.
                tolerance_us:   the synchronization tolerance, i.e. for each sample in 'reference datasource', we try to find one sample
                                in each of the other synchronized datasources that is whitin 'tolerance_us' and add it to the 'synchronized tuple'.
                                Incomplete 'synchronized tuples' are discarded.
                fifo:           For future use (live datasources)
        """
        self.platform = platform
        self.sync_labels = sync_labels
        self.interp_labels = interp_labels
        self.tolerance_us = tolerance_us
        self.fifo = fifo
        self.sync_indices = None
        self.ref_index = 0

        if not self.platform.is_live():
            self.mappings = self._synchronize_offline_sensor_data()
            self.sync_indices = self.mappings[(self.mappings != -1).all(axis=1), :]

    @property
    def ref_ds_name(self):
        return self.sync_labels[self.ref_index]

    def _synchronize_offline_sensor_data(self):
        """Synchronizes all datasources in the sync_labels list together. The first 
           datasource in the list will be considered the 'reference datasource' 
        """
        self.all_ts = [(c, self.platform[c].timestamps) for c in self.sync_labels]
        self.ref_ts = dict(self.all_ts)[self.ref_ds_name]
        mappings = []

        all_ts_tqdm = tqdm.tqdm(self.all_ts, 'Synchronizing') if self.platform.progress_bar else self.all_ts
        for name, sensor_ts in all_ts_tqdm:
            
            if sensor_ts.size == 0:
                raise RuntimeError('Sensor {} has 0 timestamps, could not \
                synchronize arrays, please exclude this label from list'
                .format(name))
            
            idx = closest_timestamps(self.ref_ts, sensor_ts, self.tolerance_us)
            mappings.append(idx)

        return np.stack(mappings, axis=1)

    def expand_wildcards(self, labels:List[str]):
        """See also: platform.expand_widlcards()""" 
        return platform_utils.expand_wildcards(labels, self.keys())

    def keys(self):
        """Returns synchronization labels and interpolation labels (implement dict API)"""
        return self.sync_labels + self.interp_labels

    def __contains__(self, key):
        return key in self.keys()

    def __len__(self):
        return self.sync_indices.shape[0] if self.sync_indices is not None else 0

    def get_single_ds(self, ds_name:str, index:Union[int, slice, Iterator[int]]) -> Sample:

        try: #is index iterable?
            l = []
            for i in index:
                l.append(self.get_single_ds(ds_name, i))
            return l
        except:
            if isinstance(index, slice):
                return self.get_single_ds(ds_name, platform_utils.slice_to_range(index, len(self)))
            elif isinstance(index, numbers.Integral):
                data_indices = self.sync_indices[index]
                try:
                    i_ds = self.sync_labels.index(ds_name)

                except:
                    try:
                        i_ds = self.interp_labels.index(ds_name)
                    except:
                        raise RuntimeError(f"Unexpected label {ds_name}")
                    return self.platform[ds_name].get_at_timestamp(self.ref_ts[data_indices[self.ref_index]])

                return self.platform[ds_name][data_indices[i_ds]]

            raise KeyError("Unsupported key type: {}".format(index))
    
    class SynchGetter:
        def __init__(self, index, synch):
            self.index = index
            self.synch = synch
            
        def __getitem__(self, ds_name):
            return self.synch.get_single_ds(ds_name, self.index)

        def __setitem__(self, key, item):
            raise RuntimeError("Immutable")

        def __delitem__(self, key):
            raise RuntimeError("Immutable")

        def __repr__(self):
            return repr(self.items())
        
        def keys(self):
            return self.synch.keys()

        def clear(self):
            raise RuntimeError("Immutable")

        def copy(self):
            return Synchronized.SynchGetter(self.index, self.synch)

        def values(self):
            return [self[k] for k in self.keys()]
        
        def items(self):
            return [(k, self[k]) for k in self.keys()]

        def len(self):
            return len(self.keys())

        def has_key(self, k):
            return k in self.keys()
        
        def __contains__(self, item):
            return item in self.values()

        def __iter__(self):
            return iter(self.values())

    def __getitem__(self, index:Union[int, slice, Iterator[int]]) ->Dict[str, Sample]:
        return Synchronized.SynchGetter(index, self)

    def timestamps(self, index:int) -> np.ndarray:
        """Returns timestamps for all synchronized datasources for a given index"""
        ts = []
        for source_i, sample_i in enumerate(self.sync_indices[index, :]):
            source = self.platform[self.sync_labels[source_i]]
            ts.append(source.timestamps[sample_i])
        return np.array(ts)

    def indices(self, index:int):
        """Returns indices for all datasources for a given synchronized index"""
        interp_indices = []
        for ds in self.interp_labels:
            interp_indices.append(self.platform[ds].to_float_index(self.ref_ts[self.sync_indices[index][self.ref_index]]))
        return list(self.sync_indices[index]) + interp_indices

    def sliced(self, intervals:List[Tuple[int, int]]):
        """Returns a Sliced instance
        
        See Also: Sliced
        """
        return Sliced(self, intervals)

    def filtered(self, indices:List[int]):
        return Filtered(self, indices)


class SynchronizedGroup(Synchronized):
    """Groups multiple synchronized platforms in a single one
        Args:
            -datasets: Can be either a list of paths for the datasets to be grouped, 
                or a single string for the path of a directory that contains the datasets that will all be grouped.
            -sync_labels: see Synchronized
            -interp_labels: see Synchronized
            -tolerance_us: see Synchronized
            -include: see Platform
            -ignore: see Platform
            -preload: If True, all platforms will be initialized and synchronized at start. 
                Good for random or non-sequential data access, but initialization is slower.
    """

    def __init__(self, datasets:Union[str,list], sync_labels:List[str]=[], interp_labels:List[str]=[], tolerance_us:Union[float, int]=None,
                     include:Optional[list]=None, ignore:Optional[list]=[], preload:bool=False, virtual_datasources_config:dict=None):
                     
        if type(datasets) == str:
            datasets = glob.glob(f"{datasets}/*")

        self.datasets = datasets
        self.sync_labels = sync_labels
        self.interp_labels = interp_labels
        self.tolerance_us = tolerance_us
        self.include = include
        self.ignore = ignore
        self.preload = preload
        self.virtual_datasources_config = virtual_datasources_config

        self.cached_synchronized = [None for d in self.datasets]
       
        # The first platform is synchronized at start
        sync0 = self.get_synchronized(0, progress_bar=True)
        self.sync_labels = sync0.sync_labels
        self.interp_labels = sync0.interp_labels
        self.tolerance_us = sync0.tolerance_us

        self._preload_lenghts()

    def _preload_lenghts(self):
    
        self.lenghts = [] 
        for dataset_index in tqdm.tqdm(range(len(self.datasets)), 'Grouping synchronized platforms'):
            try:
                if self.preload:
                    l = self._get_dataset_lenght(dataset_index)
                else:
                    l = self._get_dataset_lenght_fast(dataset_index)
            except Exception:
                l = 0
                LoggingManager.instance().warning(f"The dataset {self.datasets[dataset_index]} could not be added to the SynchronizedGroup.")

            self.lenghts.append(l)

        self.cumsum_lenghts = np.cumsum([0]+self.lenghts[:-1])

    def get_synchronized(self, dataset_index, progress_bar=False):
        if self.cached_synchronized[dataset_index] is None:
            try:
                pf = Platform(self.datasets[dataset_index], include=self.include, ignore=self.ignore, progress_bar=progress_bar, default_cache_size=0)
                if self.virtual_datasources_config is not None:
                    pf.add_virtual_datasources(self.virtual_datasources_config)
                sync = pf.synchronized(self.sync_labels, self.interp_labels, self.tolerance_us)
                self.cached_synchronized[dataset_index] = sync
            except Exception:
                LoggingManager.instance().warning(f"There is an issue with the dataset {self.datasets[dataset_index]}.")
        return self.cached_synchronized[dataset_index]

    def _get_dataset_lenght(self, dataset_index):
        sync = self.get_synchronized(dataset_index)
        if not all([ds in sync.keys() for ds in self.sync_labels]):
            return 0
        return len(sync)

    def _get_dataset_lenght_fast(self, dataset_index):
        """
        This is a faster way to get the lenght of a synchronized platform, without initalizing it.
        We directly load the timestamp.csv files and check how many matching frames are found.
        """
        # If the synchronized platform is already cached, get its lenght
        if self.cached_synchronized[dataset_index] is not None:
            return len(self.cached_synchronized[dataset_index])

        # Here, we determine for which datasources we have to load a timestamp.csv file
        datasources_with_ts = []
        for ds_name in self.sync_labels:
            datasource = self.cached_synchronized[0].platform[ds_name]
            if isinstance(datasource, virtual_datasources.VirtualDatasource):
                continue
            datasources_with_ts.append(f"{ds_name}")

        ref_ts = []
        sensor_ts = []
        mappings = []

        for ds_name in datasources_with_ts:

            # zipped case
            if os.path.exists(f"{self.datasets[dataset_index]}/{ds_name}.zip"):
                path = f"{self.datasets[dataset_index]}/{ds_name}.zip"
                with zipfile.ZipFile(path, 'r') as zf:
                    with zf.open('timestamps.csv') as stream:
                        ts = pd.read_csv(stream, delimiter=" ", dtype='u8', header=None).values[:,0]

            # unzipped case
            elif os.path.exists(f"{self.datasets[dataset_index]}/{ds_name}/timestamps.csv"):
                path = f"{self.datasets[dataset_index]}/{ds_name}/timestamps.csv"
                with open(path) as stream:
                    ts = pd.read_csv(stream, delimiter=" ", dtype='u8', header=None).values[:,0]
            else:
                return 0

            if ref_ts == []:
                ref_ts = ts

            idx = closest_timestamps(ref_ts, ts, self.tolerance_us)
            mappings.append(idx)
        
        mappings = np.stack(mappings, axis=1)
        mappings = mappings[(mappings != -1).all(axis=1), :]

        return mappings.shape[0]

    def __len__(self):
        return sum(self.lenghts)

    @property
    def platform(self):
        return self.cached_synchronized[0].platform
 
    def timestamps(self, index:int):
        raise RuntimeError('not implemented!')

    def indices(self, index:int):
        raise RuntimeError('not implemented!')

    def sliced(self, intervals:List[Tuple[int, int]]):
        raise RuntimeError('not implemented!')

    def filtered(self, indices:List[int]):
        raise RuntimeError('not implemented!')

    def keys(self):
        return self.cached_synchronized.keys()

    def _group_index(self, index:int):
        """For a given index from 0 to the summed lenght of all grouped datasets,
        finds in which dataset to get the samples and what is the corresponding index in this dataset.
        Ex: If we grouped two datasets of lenght 1000 each, then _group_index(1500) would return 
            the pair 1, 500. So, the samples are at index 500 in the second (dataset_index=1) dataset.
        """
        if index > len(self) or index < 0:
            raise IndexError("Index out of bounds")
        
        cumsum_index = index-self.cumsum_lenghts
        dataset_index = np.where(cumsum_index<self.lenghts)[0][0]
        idx = cumsum_index[dataset_index] #find the index in the specific dataset
        return dataset_index, idx

    def __getitem__(self, index:Union[int, slice, Iterator[int]]) ->Dict[str, Sample]:

        list_index = []
        try:
            for i in index:
                list_index.append(i)
        except:
            if isinstance(index, slice):
                list_index = list(platform_utils.slice_to_range(index, len(self)))
            if isinstance(index, numbers.Integral):
                list_index = [index]
        
        merged = None
        for i in list_index:
            dataset_index, idx = self._group_index(i)
            samples = self.get_synchronized(dataset_index)[idx]
            if merged is None:
                merged = {k:[v] for k,v in samples.items()}
            else:
                for k,v in merged.items():
                    v.append(samples[k])
        return merged



class Filtered:

    def __init__(self, synchronized:Synchronized, indices:List[int]):
        self.synchronized = synchronized
        self.indices = indices

    @property
    def platform(self):
        return self.synchronized.platform
        
    def expand_wildcards(self, labels:List[str]):
        """Wraps Synchronized.expand_widlcards()
        """ 
        return self.synchronized.expand_wildcards(labels)

    def keys(self):
        """Wraps Synchronized.keys()"""
        return self.synchronized.keys()

    def __contains__(self, key):
        """Wraps Synchronized.__contains__()"""
        return self.synchronized.__contains__(key)

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, index:Any) ->Dict[str, Sample]:
        """Implements the '[]' API, using the intervals

        Args:
            index: Will index in this view's intervals, and then call Synchronized.__getitem__() 
        Returns:
            See Synchronized.__getitem__()

        """
        return self.synchronized[self.indices[index]]
        

class Sliced:
    """Creates a view on a synchronized dataset using a list of intervals"""

    def __init__(self, synchronized:Synchronized, intervals:List[Tuple[int, int]], stride:int = 1):
        """Constructor

            Args:
                synchronized:   The 'Synchronized' instance we want a view on.
                intervals:      An list of (open, open) intervals of interest over synchronized's domain. 
                                The list does not need to be ordered, and can index a synchronized index more than once.
                                For example, with a 'stride' of 1, the list [(5, 7), (12, 15), (15, 13)] will expand to indices [5,6,7, 12,13,14,15, 15,14,13].
                stride:         the stride (jump between consecutive frames)
                              
        """
        self.synchronized = synchronized
        self.intervals = intervals
        self.stride = stride
        self.indices = []
        n = len(synchronized)
        for s,e in intervals:
            expanded = []
            if s>=0 and e>=0 and s > e: #inverted
                expanded = [i for i in range(e+1, s+1, self.stride)][::-1]
            else:
                expanded = [i for i in range(s, e+1, stride)]

            self.indices.extend(expanded)
    
    @property
    def platform(self):
        return self.synchronized.platform
        
    def expand_wildcards(self, labels:List[str]):
        """Wraps Synchronized.expand_widlcards()
        """ 
        return self.synchronized.expand_wildcards(labels)

    def keys(self):
        """Wraps Synchronized.keys()"""
        return self.synchronized.keys()

    def __contains__(self, key):
        """Wraps Synchronized.__contains__()"""
        return self.synchronized.__contains__(key)

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, index:Any) ->Dict[str, Sample]:
        """Implements the '[]' API, using the intervals

        Args:
            index: Will index in this view's intervals, and then call Synchronized.__getitem__() 
        Returns:
            See Synchronized.__getitem__()

        """
        return self.synchronized[self.indices[index]]


class Sensors:
    """The collection of Sensor instances in a platform"""
    def __init__(self, pf:'Platform', yml:dict):
        """Constructor

           Args:
            platform: the platform that holds this Sensors instance
            yml: the YAML database
        """
        self.platform = pf
        self._sensors = {}
        self._ordered_names = []
        self._egomotion_provider = None

        yml_sensor_items = [s for s in yml.items() if s[0] not in ['ignore','virtual_datasources','synchronization']]
        yml_items_tqdm = tqdm.tqdm(yml_sensor_items, 'Loading sensors') if self.platform.progress_bar else yml_sensor_items
        for name, value in yml_items_tqdm:

            sensor_type, _ = platform_utils.parse_sensor_name(name)
            self._sensors[name] = SENSOR_FACTORY.get(sensor_type, Sensor)(name, self.platform)
            self._ordered_names.append(name)

            self._load_offline_datasources(name)

            if 'orientation' in value:
                m = np.array(value['orientation'], dtype = 'f4')
                if m.shape != (3,3):
                    LoggingManager.instance().warning('Ignoring orientation for sensor {}: {} is \
                    not a 3 x 3'.format(name, str(m)))
                else:
                    self._sensors[name].orientation = m

            if 'intrinsics' in value:
                self._load_intrinsics(name, value['intrinsics'])

            if 'extrinsics' in value:
                self._load_extrinsics(name, value['extrinsics'])

        self.__add_egomotion_provider()

    def __add_egomotion_provider(self):
        for name, sensor in self.items():
            try:
                provider = sensor.create_egomotion_provider()
            except:
                LoggingManager.instance().warning(f"The 'egomotion_provider' for sensor name {name} could not be created.")

            if provider is not None:
                if self._egomotion_provider is not None:
                    LoggingManager.instance().warning(f"Another 'egomotion_provider' found for sensor name {name}, ignoring it.")
                else:
                    self._egomotion_provider = provider
    
    def _load_intrinsics(self, name, intrinsics_config):
        """Intrinsics config can be a string or a dict so complete the path,
           only if the intrinsics config is a string.
        """
        if isinstance(intrinsics_config, six.string_types):
            intrinsics_config = self.platform.to_nas_path(intrinsics_config)
        self._sensors[name].load_intrinsics(intrinsics_config)

    def _load_extrinsics(self, name, extrinsics_folder):
        """ Extrinsics files are pkl files that contain a 4x4 numpy array that
            allows to project 3d point(s) form a sensor's referential to another
            sensor's referential. The expected file format is
            "{}-{}.pkl".format(source_sensor_name, destination_sensor_name).
            For example, file "eagle_bcc-flir_tfl.pkl" would contain a 4x4 numpy
            array that allows to project an eagle_bcc's point cloud in flir_tfl's
            referential.

            Note it will also try to find
            "{}-{}.pkl".format(destination_sensor_name, source_sensor_name) and
            use that transform's inverse
        """
        self._sensors[name].load_extrinsics(extrinsics_folder)

    def _load_offline_datasources(self, name:str):

        archives = glob.glob(os.path.join(self.platform.dataset, name + '_*'))
        for archive in [os.path.basename(f) for f in archives]:

            if archive.endswith('.zip'):
                # remove the .zip and extracts the 'datasource' suffix:
                ds_name = os.path.splitext(archive)[0].split('_')[-1]#.split('-')[0]
                try:
                    ds = ZipFileSource(os.path.join(self.platform.dataset, archive))
                    self._sensors[name].add_datasource(ds, ds_name, cache_size = self.platform.default_cache_size)
                except:
                    LoggingManager.instance().warning(f'Zip file for {name}_{ds_name} could not be loaded.')
                    continue

            else:
                ds_name = archive.split('_')[-1]#.split('-')[0]
                try:
                    ds = DirSource(os.path.join(self.platform.dataset, archive))
                    self._sensors[name].add_datasource(ds, ds_name, cache_size = self.platform.default_cache_size)
                except:
                    LoggingManager.instance().warning(f'Data directory for {name}_{ds_name} could not be loaded.')
                    continue

    def override_sensor_extrinsics(self, name:str, extrinsics_folder:str):
        """Override the extrinsics that we loaded from the platfrom yml.

        Args:
            extrinsics_folder: The extrinsics folder
        """
        self._load_extrinsics(name, extrinsics_folder)

    def override_all_extrinsics(self, extrinsics_folder:str):
        """Override the extrinsics that we loaded from the platfrom yml.

        Only the sensor for which the extrinsics were previously loaded
        will be overriden.

        Args:
            extrinsics_folder: The extrinsics folder
        """
        for name, sensor in self.items():
            if sensor.extrinsics is not None:
                self._load_extrinsics(name, extrinsics_folder)

    def keys(self):
        """Implement dict API
            Returns:
                The list of sensor names, e.g. ['pixell_tfc', 'flir_tfc', ...]
        """
        return self._sensors.keys()

    def items(self):
        """Implement dict API"""
        return self._sensors.items()

    def values(self):
        """Implement dict API"""
        return self._sensors.values()

    def __contains__(self, key):
        return key in self._sensors

    def __len__(self):
        return len(self._sensors)

    def __getitem__(self, label:str):
        """Implement '[]' API"""
        return self._sensors[label]

    def __iter__(self):
        return iter(self._sensors)


import numbers
import numpy as np

def from_float_index(float_index):
    i = int(np.floor(float_index))
    t = float_index - i
    return i, t

def _linear_interpolate_helper(t, from_val, to_val):
    return t * to_val + (1 - t) * from_val

def _angles_linear_interpolate_helper(t, from_val, to_val):
    if from_val-to_val>np.pi:
        return t * (to_val+2.0*np.pi) + (1 - t) * from_val
    elif from_val-to_val<-np.pi:
        return t * to_val + (1 - t) * (from_val+2.0*np.pi)
    else:
        return t * to_val + (1 - t) * from_val


def linear_ndarray_interpolator(datasource, float_index):
    i, t = from_float_index(float_index)
    if i+1 != len(datasource):
        raw_from, raw_to = datasource[i].raw, datasource[i+1].raw
    else:
        raw_from, raw_to = datasource[i].raw, datasource[i].raw
    if raw_from.dtype.names is None:
        return _linear_interpolate_helper(t, raw_from, raw_to)
    
    result = np.zeros_like(raw_from)
    for name in raw_from.dtype.names:
        result[name] = _linear_interpolate_helper(t, raw_from[name], raw_to[name])
    return result


def euler_imu_linear_ndarray_interpolator(datasource, float_index):
    i, t = from_float_index(float_index)
    if i+1 != len(datasource):
        raw_from, raw_to = datasource[i].raw, datasource[i+1].raw
    else:
        raw_from, raw_to = datasource[i].raw, datasource[i].raw

    if raw_from.dtype.names is None:
        return _linear_interpolate_helper(t, raw_from, raw_to)
    
    result = np.zeros_like(raw_from)
    for name in raw_from.dtype.names:
        if name in ['roll', 'pitch', 'yaw']:
            result[name] = _angles_linear_interpolate_helper(t, raw_from[name], raw_to[name])
        else:
            result[name] = _linear_interpolate_helper(t, raw_from[name], raw_to[name])
    return result


def linear_dict_of_float_interpolator(datasource, float_index):
    i, t = from_float_index(float_index)
    raw_from, raw_to = datasource[i].raw, datasource[i+1].raw
    interp_dict = raw_from

    for k,value_from in raw_from.items():
        value_to = raw_to[k]
        if isinstance(value_from, list):
            interp_dict[k] = _linear_interpolate_helper(t, np.array(value_from), np.array(value_to)).tolist()
        elif isinstance(value_from, numbers.Real):
            interp_dict[k] = type(value_from)(_linear_interpolate_helper(t, value_from, value_to))
    return interp_dict

def floor_interpolator(datasource, float_index):
    i, _ = from_float_index(float_index)
    return datasource[i].raw

def ceil_interpolator(datasource, float_index):
    i, _ = from_float_index(float_index)
    return datasource[i+1].raw

def nearest_interpolator(datasource, float_index):
    return datasource[round(float_index)].raw



import cv2
import os
import numpy as np
import pickle
import re
import six

def load_files_from_folder(folder, pattern, sort:bool=False, return_keys:bool=False):
    """Load files from a folder and sort them according to a regex pattern"""
    if return_keys:
        assert sort, 'sort must be True when return_keys is True'

    files = os.listdir(folder)
    regex = re.compile(pattern)
    output = []
    for f in files:
        match = regex.match(f)
        if match:
            groups = match.groups()
            if sort and not groups:
                raise ValueError('no groups')
            if sort and groups:
                sample = (int(groups[0]), os.path.join(folder, f))
            else:
                sample = os.path.join(folder, f)

            output.append(sample)
    if sort:
        output.sort()
        keys = [s[0] for s in output]
        output = [s[1] for s in output]
    if return_keys and sort:
        return output, keys
    else:
        return output


def txt_loader(fileobj):
    """Bytes data file loader. To be used with FileSource.

    Arguments:
        fileobj {string,file} -- A file object or a string

    Returns:
        bytes
    """
    if isinstance(fileobj, bytes):
        data = fileobj.decode('utf-8')
    elif isinstance(fileobj, six.string_types):
        with open(fileobj, 'rb') as f:
            data = f.read().decode('utf-8')
    elif hasattr(fileobj, 'read'):
        data = fileobj.read().decode('utf-8')
    else:
        raise ValueError('fileobj is not a filename or a file object')
    return data

def pickle_loader(fileobj):
    """Pickle data file loader. To be used with FileSource.

    Arguments:
        fileobj {string,file} -- A file object or a string

    Returns:
        object -- Unpickled object
    """
    if isinstance(fileobj, bytes):
        data = pickle.loads(fileobj, encoding="latin1")
    elif isinstance(fileobj, six.string_types):
        with open(fileobj, 'rb') as f:
            data = pickle.load(f, encoding="latin1")
    elif hasattr(fileobj, 'read'):
        data = pickle.load(fileobj, encoding="latin1")
    else:
        raise ValueError('fileobj is not a filename or a file object')
    return data

def image_loader(fileobj):
    """Image loader to be used with FileSource

    Arguments:
        fileobj {string} -- Image filename

    Returns:
        np.ndarray -- Image
    """
    if isinstance(fileobj, six.string_types):
        return cv2.imread(fileobj, cv2.IMREAD_COLOR)[..., ::-1] #bgr->rgb
    elif isinstance(fileobj, bytes):
        byte_arr = bytearray(fileobj)
    else:
        byte_arr = bytearray(fileobj.read())
    
    return cv2.imdecode(np.asarray(byte_arr, dtype=np.uint8), cv2.IMREAD_COLOR)[..., ::-1] #bgr->rgb



from pioneer.das.api.datasources.datasource import DataSource
from pioneer.das.api.datasources.abstract_datasource import AbstractDatasource
from pioneer.das.api.datasources.datasource_wrapper import DatasourceWrapper
from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource


from pioneer.das.api.interpolators import nearest_interpolator
from pioneer.das.api.samples.annotations.box_3d import Box3d
from pioneer.das.api.samples.annotations.box_2d import Box2d
from pioneer.das.api.samples.annotations.poly_2d import Poly2d
from pioneer.das.api.samples.annotations.seg_2d import Seg2d
from pioneer.das.api.samples.annotations.seg_2d_image import Seg2dImage
from pioneer.das.api.samples.annotations.seg_3d import Seg3d
from pioneer.das.api.samples.annotations.lane import Lane


ANNOTATIONS_FACTORY = {
    'box3d': (Box3d, nearest_interpolator),
    'box2d': (Box2d, nearest_interpolator),
    'poly2d': (Poly2d, nearest_interpolator),
    'seg2d': (Seg2d, nearest_interpolator),
    'seg2dimg': (Seg2dImage, nearest_interpolator),
    'seg3d': (Seg3d, nearest_interpolator),
    'lane': (Lane, nearest_interpolator),
}


from pioneer.das.api                import datatypes
from pioneer.das.api.samples.sample import Sample

import copy
import cv2
import numpy as np

class Seg2dImage(Sample):

    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super(Seg2dImage, self).__init__(index, datasource, virtual_raw, virtual_ts)

    def seg2d(self):
        category_img = copy.deepcopy(self.raw)

        seg2d = {'data': np.empty(0, dtype=datatypes.seg2d())}

        for category_number in np.unique(category_img):

            category_array = np.zeros_like(category_img)
            mask = np.where(category_img == category_number)
            category_array[mask] = 1 # confidence 1 is forced here

            seg2d['data'] = np.append(seg2d['data'], np.array([(category_array, category_number)], dtype=datatypes.seg2d()))

        return seg2d


    def poly2d(self, confidence_threshold=0.5):

        seg2d = self.seg2d()

        poly2d = np.empty(0, dtype=datatypes.poly2d())
        for seg_data in seg2d['data']:
            mask = (seg_data['confidences']>0).astype(np.uint8)
            if mask.max()==0:
                continue
            polys,_ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            polys = np.array(polys)
            for poly in polys:
                poly = poly[:,0,:]
                poly2d = np.append(poly2d, np.array([(poly, seg_data['classes'],0,0)], dtype=datatypes.poly2d()))

        return poly2d



from pioneer.das.api.datasources import AbstractDatasource

from typing import List, Optional, Callable, Dict, Any

import numpy as np

class VirtualDatasource(AbstractDatasource):
    def __init__(self, ds_type:str, dependencies:List[str], callback:Optional[Callable[[Dict[str, AbstractDatasource], Any], 'Sample']], timestamps_source = None):
        """Constructor
        Args:
            ds_type:        the datasource type, e.g. 'ech-bbox'
            dependencies:   a list of datasource names used to compute this datasources, e.g. ['eagle_tfc_ech', 'flir_tfc_img', ...].
            callback:       a method that will be called with arguments\:
                                'datasources'\: a dict with dependencies as keys and corresponding datasources as values
                                'key'\:  the argument passed to the virtual datasource's __getitem__().
            timestamps_source: datasource used as timestamps, If None first one dependencies used
        """
        super(VirtualDatasource, self).__init__(None, ds_type)
        self.ds_type = ds_type
        self.dependencies = dependencies
        self.datasources = None #will be configured during call to _set_sensor 
        self.callback = callback
        self.timestamps_source = timestamps_source

    def invalidate_caches(self):
        #No cache implemented for virtual datasources
        pass

    def _set_sensor(self, sensor:'Sensor'):
        """Sets this datasource's sensor"""
        self.sensor = sensor
        self.datasources = {dep:self.sensor.platform[dep] for dep in self.dependencies}

    def get_first_datasource(self) -> AbstractDatasource:
        if self.timestamps_source is None:
            for k in self.datasources:
                return self.datasources[k] #length of the first ds
        else:
            return self.datasources[self.timestamps_source]
        raise RuntimeError('no datasources')

    @property
    def timestamps(self) -> np.ndarray:
        return self.get_first_datasource().timestamps

    def get_at_timestamp(self, timestamp:int):
        """Default implementation"""

        if hasattr(self.callback, 'get_at_timestamp'):
            return self.callback.get_at_timestamp(self.datasources, timestamp)

        float_index = self.get_first_datasource().to_float_index(timestamp)
        return self[int(np.floor(float_index))]

    def __getitem__(self, key):
        return self.callback(self.datasources, key)

    @property
    def label(self):
        return "{}_{}".format(self.sensor.name, self.ds_type)



from pioneer.das.api.sensors.sensor import Sensor
from pioneer.das.api.sensors.sensor3d import Sensor3D
from pioneer.das.api.sensors.lidar import Lidar
from pioneer.das.api.sensors.radar import Radar
from pioneer.das.api.sensors.pixell import Pixell
from pioneer.das.api.sensors.motor_lidar import MotorLidar
from pioneer.das.api.sensors.camera import Camera
from pioneer.das.api.sensors.imu_sbg_ekinox import ImuSbgEkinox
from pioneer.das.api.sensors.carla_imu import CarlaIMU 
from pioneer.das.api.sensors.lcax import LCAx
from pioneer.das.api.sensors.lca3 import LCA3
from pioneer.das.api.sensors.gps_vaya import GPSVaya



SENSOR_FACTORY = {  
    'lca2': LCAx,
    'pixell': Pixell,
    'lca3': LCA3,
    'eagle': LCA3,
    'flir': Camera,
    'camera': Camera,
    'sbgekinox': ImuSbgEkinox,
    'vlp16': Lidar,
    'ouster64': MotorLidar,
    'radarTI': Radar,
    'radarConti': Radar,
    'webcam': Camera,
    'carlaimu': CarlaIMU,
    'leddar': LCAx,
    'lidar': Lidar,
    'gpsvaya': GPSVaya,
}


from pioneer.common import banks, clouds, images, plane
from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.samples.sample import Sample

import copy
import numpy as np
import warnings
warnings.simplefilter('once', DeprecationWarning)


class Echo(Sample):
    """Sample from a single data package provided by a LCAx sensor. 
        See pioneer.common.clouds.to_echo_package() to create a similar data package from scratch.
    """

    def __init__(self, index, datasource, virtual_raw=None, virtual_ts=None):
        super().__init__(index, datasource, virtual_raw, virtual_ts)
        self._mask = None

    @property
    def raw(self):
        if self._raw is None:
            r = super().raw
            r['das.sample'] = self
            
            try:
                if r['data']['timestamps'][-1] == 0:
                    cfg = self.datasource.sensor['cfg'].get_at_timestamp(
                        self.timestamp).raw
                    if self.datasource.sensor.specs is None:
                        self.datasource.sensor.specs = banks.extract_specs(
                            lambda n: cfg[n])

                    try:
                        if self.datasource.sensor.modules_angles is None:
                            self.datasource.sensor.modules_angles = banks.extract_intrinsics_modules_angles(
                                lambda n: cfg[n])
                    except:
                        LoggingManager.instance().warning(
                            "Sensor {} has no modules angles, or can not read modules angles from the sensor intrinsic calibratrion".format(self.datasource.sensor.name))

                    banks.add_timestamp_offsets(r, self.datasource.sensor.name, self.datasource.sensor.specs,
                                                int(cfg['ID_ACCUMULATION_EXP']), int(cfg['ID_OVERSAMPLING_EXP']), int(cfg['ID_BASE_POINT_COUNT']))
            except:
                pass

            self._raw = r

        return self._raw

    @property
    def specs(self):
        # override the sensor specs if they are present in the YAML config file
        sensor_specs = self.datasource.sensor.specs
        if sensor_specs is not None:
            return sensor_specs
        return {k: self.raw[k] for k in ['v', 'h', 'v_fov', 'h_fov']}

    @property
    def v(self):
        return self.raw['v']

    @property
    def h(self):
        return self.raw['h']

    @property
    def size(self) -> int:
        return self.indices.size

    @property
    def mask(self):
        if self._mask is None:
            data = self.raw['data']
            config = self.datasource.sensor.config

            vmask = np.bitwise_and(data['flags'], 0x01).astype(bool)  # keep valid echoes only
            fmask = np.isin(
                data['flags'], config['reject_flags'], invert=True) & vmask

            def get_mask(slices, values):
                dmask = np.full_like(fmask, True)
                for s, e in slices:
                    dmask = dmask & ~((values >= s) & (values <= e))
                return dmask

            self._mask = fmask & get_mask(config['dist_reject_intervals'], data['distances']) & get_mask(
                config['amp_reject_intervals'], data['amplitudes'])

        return self._mask

    @property
    def masked(self):
        r = dict(self.raw)
        r['data'] = self.data
        return r

    @property
    def data(self):
        return self.raw['data'][self.mask]

    @property
    def indices(self):
        return self.data['indices']

    @property
    def timestamps(self):
        try:  # if relative timestamps are converted to global, there will be a key 'timestamps' in the raw
            t_ = self.raw['timestamps'][self.mask]
        except:  # otherwise global timestamps are considered to be directly in the data
            t_ = self.raw['data']['timestamps'][self.mask]
        return t_

    @property
    def distances(self):
        d = self.data['distances']
        s = self.datasource.sensor
        d = s.apply_distance_corrections(self.timestamp, self.indices, d)
        return d

    @property
    def amplitudes(self):
        return self.data['amplitudes']

    @property
    def flags(self):
        return self.data['flags']

    def cache(self):
        return self.datasource.sensor.cache(self.specs)

    def get_point_cloud(self, referential: str = None, ignore_orientation: bool = False, undistort: bool = False, reference_ts: int = -1,
                    dtype: np.dtype = np.float64):
        """Compute a 3D point cloud from raw data

        Args:
            referential: The target sensor referential or full datasource name
            ignore_orientation: Ignore the source sensor orientation (default: {False})
            undistort: Apply motion compensation to 3d points.
            reference_ts: (only used if referential == 'world' and/or undistort == True), refer to compute_transform()'s documentation
            dtype: the output numpy data type
        """

        pts_Local = self.datasource.sensor.get_corrected_cloud(
            self.timestamp, self.cache(), 'point_cloud', self.indices, self.distances, None, dtype)

        if undistort:
            to_world = referential == 'world'
            self.undistort_points(pts_Local, self.timestamps, reference_ts, to_world, dtype=dtype)
            if to_world:
                return pts_Local  # note that in that case, orientation has to be ignored

        return self.transform(pts_Local, referential, ignore_orientation, reference_ts, dtype=dtype)

    def point_cloud(self, referential=None, ignore_orientation=False, undistort=False, reference_ts=-1, dtype=np.float64):
        warnings.warn("Echo.point_cloud() is deprecated. Use Echo.get_point_cloud() instead.", DeprecationWarning)
        return self.get_point_cloud(referential, ignore_orientation, undistort, reference_ts, dtype)

    def get_cloud(self, referential: str = None, ignore_orientation: bool = False, undistort: bool = False, reference_ts: int = -1, dtype: np.dtype = np.float64):
        points, amplitudes, triangles = self.quad_cloud(
            referential, ignore_orientation, undistort, reference_ts, dtype)
        return points, amplitudes, triangles.reshape(-1, 3)

    def quad_cloud(self, referential=None, ignore_orientation=False, undistort=False, reference_ts=-1, dtype=np.float64):
        """Compute a 3d surface cloud from raw data (quads made of 2 triangles)

        Args:
            referential: The target sensor referential or full datasource name
            ignore_orientation: Ignore the source sensor orientation (default: {False})
            undistort: Apply motion compensation to 3d points.
            reference_ts: (only used if referential == 'world' and/or undistort == True), 
                          refer to compute_transform()'s documentation
            dtype: the output numpy data type
        """

        pts_Local, quad_amplitudes, quad_indices = self.datasource.sensor.get_corrected_cloud(
            self.timestamp, self.cache(), 'quad_cloud', self.indices, self.distances, self.amplitudes, dtype)

        if undistort:
            sn = self.indices.shape[0]

            # four points per quad, 1 different direction per point, same distance for each
            to_world = referential == 'world'
            self.undistort_points(pts_Local, np.tile(self.timestamps, 4), reference_ts, to_world, dtype=dtype)
            if to_world:
                # note that in that case, orientation has to be ignored
                return pts_Local, quad_amplitudes, quad_indices

        pts_Ref = self.transform(
            pts_Local, referential, ignore_orientation, reference_ts, dtype=dtype)
        return pts_Ref, quad_amplitudes, quad_indices

    def amplitude_img(self, options='max_amplitude', dtype=np.float32, extrema=None):

        if extrema is None:
            extrema = self.datasource.sensor.config['extrema_amp']

        if options == 'max_amplitude':
            img = images.extrema_image(
                self.v, self.h, self.data, sort_field='amplitudes', sort_direction=-1, dtype=dtype, extrema=extrema)
        elif options == 'min_distance':
            _, others = images.extrema_image(self.v, self.h, self.data, sort_field='distances', sort_direction=1,
                                             other_fields=['amplitudes'], dtype=dtype, extrema=extrema)
            img = others['amplitudes']

        elif options == 'amplitudes_sum':
            img = images.accumulation_image(
                self.v, self.h, indices=self.data['indices'], weights=self.data['amplitudes'], dtype=dtype)
        else:
            raise ValueError('Invalid options: {}'.format(options))

        return self.transform_image(img)

    def distance_img(self, options='min_distance', dtype=np.float32, extrema=None):

        if extrema is None:
            extrema = self.datasource.sensor.config['extrema_dist']

        if options == 'max_amplitude':
            _, others = images.extrema_image(self.v, self.h, self.data, sort_field='amplitudes',
                                             sort_direction=-1, other_fields=['distances'], dtype=dtype, extrema=extrema)
            img = others['distances']

        elif options == 'min_distance':
            img = images.extrema_image(
                self.v, self.h, self.data, sort_field='distances', sort_direction=1, dtype=dtype, extrema=extrema)
        elif options == 'distances_sum':
            img = images.accumulation_image(
                self.v, self.h, indices=self.data['indices'], weights=self.data['distances'], dtype=dtype)
        else:
            raise ValueError('Invalid options: {}'.format(options))

        return self.transform_image(img)

    def other_field_img(self, field, dtype=np.float32):
        amp, others = images.extrema_image(self.v, self.h, self.data, sort_field='amplitudes',
                                           sort_direction=-1, other_fields=[field], dtype=dtype, extrema=None)
        img = others[field]

        return self.transform_image(img)

    @property
    def coords_img_tf(self):
        if not hasattr(self, '_coords_img_tf'):
            vv, hh = np.mgrid[0:self.v, 0:self.h]
            coords_img = np.stack(
                (vv, hh, np.arange(0, self.v*self.h).reshape(self.v, self.h)), axis=2)
            self._coords_img_tf = self.transform_image(coords_img)
        return self._coords_img_tf

    def image_coord_to_channel_index(self, row, col):
        return self.coords_img_tf[row, col, 2]

    def channel_index_to_image_coord(self, index):
        return np.argwhere(self.coords_img_tf[..., 2] == index)[0]

    def image_stack(self, field: str, number_images: int = 3, options='min_distance', dtype=np.float32):
        """Returns multiple images to account for the multiple echoes per channel"""
        image_stack = np.zeros((number_images, self.v, self.h))
        remaining_data = copy.deepcopy(self.data)
        for i in range(number_images):
            if remaining_data.size == 0:
                break
            if options == 'min_distance':
                mask = images.echoes_visibility_mask(remaining_data)
            elif options == 'max_amplitude':
                mask = images.maximum_amplitude_mask(remaining_data)

            image_stack[i] = self.transform_image(images.extrema_image(self.v, self.h, remaining_data[mask],
                                                                       sort_direction=1, other_fields=[field], dtype=dtype)[1][field])
            remaining_data = remaining_data[~mask]
        return image_stack

    def clip_to_fov_mask(self, pts: np.ndarray) -> np.ndarray:

        lcax = self.datasource.sensor
        specs = self.specs

        if lcax.angle_chart:
            cache = self.cache()
            correct_v_angles = lcax.get_corrected_projection_data(
                self.timestamp, cache, 'angles')

            v_cell_size, h_cell_size = clouds.v_h_cell_size_rad(specs)

            planes = clouds.frustrum_planes(clouds.custom_frustrum_directions(
                correct_v_angles, v_cell_size, h_cell_size, dtype=np.float64))
        else:
            planes = clouds.frustrum_planes(clouds.frustrum_directions(
                specs['v_fov'], specs['h_fov'], dtype=np.float64))

        if self.orientation is not None:
            planes[:, :3] = (self.orientation @ (planes[:, :3].T)).T

        return plane.plane_test(planes[0], pts)\
            & plane.plane_test(planes[1], pts)\
            & plane.plane_test(planes[2], pts)\
            & plane.plane_test(planes[3], pts)

    def get_pulses(self, trace_ds_type: str, pulse_sample_size: int = 10, trace_processing=None, return_distance_scaling: bool = False):
        """Returns the corresponding pulses for every echoes from a trace datasource.

            Args:
                trace_ds_type: (str) the type of trace datasource (ex: 'trr' or 'ftrr')
                pulse_sample_size: (int) The number of data points to gather before and after each echo in the waveforms.
                    For example, with pulse_sample_size=10, the pulses will be 21 points, because the highest point is
                    taken in addition to the 10 points before and the 10 points after.
                trace_processing: (callable) function applied to the traces.
                return_distance_scaling: (bool) If True, also return the distance (meters) between trace data points.
        """
        trace_sample = self.datasource.sensor[trace_ds_type].get_at_timestamp(
            self.timestamp)
        traces = trace_sample.raw if trace_processing is None else trace_sample.processed(
            trace_processing)
        if 'high' in traces:
            traces = traces['high']

        full_traces = traces['data'][self.indices]

        if isinstance(traces['time_base_delays'], float):
            time_base_delays = traces['time_base_delays'] 
        else: 
            time_base_delays = traces['time_base_delays'][self.indices]

        echoes_positions_in_traces = (
            (self.distances - time_base_delays)/traces['distance_scaling']).astype(int)
        ind = np.indices(echoes_positions_in_traces.shape)
        padded_traces = np.pad(
            full_traces, ((0, 0), (pulse_sample_size, pulse_sample_size+1)))

        pulses = np.vstack([padded_traces[ind, echoes_positions_in_traces+ind_pulse+pulse_sample_size]
                            for ind_pulse in np.arange(-pulse_sample_size, pulse_sample_size+1, dtype=int)]).T

        if return_distance_scaling:
            return pulses, traces['distance_scaling']
        return pulses

    def get_signal_to_noise(self, trace_ds_type: str):
        traces = self.datasource.sensor[trace_ds_type].get_at_timestamp(
            self.timestamp).raw
        if 'high' in traces:
            traces = traces['high']
        traces_zeroed = traces['data'] - \
            np.mean(traces['data'], axis=1)[:, None]
        noise = np.std(traces_zeroed, axis=1)[self.indices]
        return np.log10(self.amplitudes**2/noise)+1




LANE_TYPES = {
    0: {
        'name': 'curb',
        'color': (200, 200, 200),
        'double': False,
        'dashed': False,
    },

    1: {
        'name': 'solid_yellow',
        'color': (220, 190, 40),
        'double': False,
        'dashed': False,
    },

    2: {
        'name': 'dashed_yellow',
        'color': (220, 190, 40),
        'double': False,
        'dashed': True,
    },

    3: {
        'name': 'double_solid_yellow',
        'color': (220, 190, 40),
        'double': True,
        'dashed': [False, False],
    },

    4: {
        'name': 'dashed_solid_yellow',
        'color': (220, 190, 40),
        'double': True,
        'dashed': [True, False],
    },

    5: {
        'name': 'solid_dashed_yellow',
        'color': (220, 190, 40),
        'double': True,
        'dashed': [False, True],
    },

    6: {
        'name': 'solid_white',
        'color': (240, 240, 240),
        'double': False,
        'dashed': False,
    },

    7: {
        'name': 'dashed_white',
        'color': (240, 240, 240),
        'double': False,
        'dashed': True,
    },

    8: {
        'name': 'double_solid_white',
        'color': (240, 240, 240),
        'double': True,
        'dashed': [False, False],
    },

    9: {
        'name': 'dashed_solid_white',
        'color': (240, 240, 240),
        'double': True,
        'dashed': [True, False],
    },

    10: {
        'name': 'solid_dashed_white',
        'color': (240, 240, 240),
        'double': True,
        'dashed': [False, True],
    },

    100: {
        'name': 'solid_blue',
        'color': (0x1f, 0x77, 0xb4),
        'double': False,
        'dashed': False,
    },

    101: {
        'name': 'solid_orange',
        'color': (0xff, 0x7f, 0x0e),
        'double': False,
        'dashed': False,
    },

    102: {
        'name': 'solid_green',
        'color': (0x2c, 0xa0, 0x2c),
        'double': False,
        'dashed': False,
    },

    103: {
        'name': 'solid_red',
        'color': (0xd6, 0x27, 0x28),
        'double': False,
        'dashed': False,
    },

    104: {
        'name': 'solid_purple',
        'color': (0x94, 0x67, 0xbd),
        'double': False,
        'dashed': False,
    },

    105: {
        'name': 'solid_brown',
        'color': (0x8c, 0x56, 0x4b),
        'double': False,
        'dashed': False,
    },

    106: {
        'name': 'solid_pink',
        'color': (0xe3, 0x77, 0xc2),
        'double': False,
        'dashed': False,
    },

    107: {
        'name': 'solid_gray',
        'color': (0x7f, 0x7f, 0x7f),
        'double': False,
        'dashed': False,
    },

    108: {
        'name': 'solid_olive',
        'color': (0xbc, 0xbd, 0x22),
        'double': False,
        'dashed': False,
    },

    109: {
        'name': 'solid_cyan',
        'color': (0x17, 0xbe, 0xcf),
        'double': False,
        'dashed': False,
    },

}



def get_number(name):
    for number in LANE_TYPES:
        if LANE_TYPES[number]['name'] == name:
            return number
    


from pioneer.das.api.egomotion.egomotion_provider  import EgomotionProvider
from pioneer.das.api.datasources                   import DatasourceWrapper

import numpy as np



class VayaEgomotionProvider(EgomotionProvider):
    """Ego Motion provider from Vayadrive exported datasets"""

    def __init__(self, referential_name:str, matrix_datasource:DatasourceWrapper):
        super().__init__(referential_name)
        self.matrix_datasource = matrix_datasource

    def get_transform(self, timestamp:int) -> np.ndarray:
        return self.matrix_datasource.get_at_timestamp(timestamp).raw

    def get_timestamps(self) -> np.ndarray:
        return self.matrix_datasource.timestamps
        


from pioneer.common import clouds, platform
from pioneer.common.trace_processing import Binning, Decimate, TraceProcessingCollection, ZeroBaseline
from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource
from pioneer.das.api.samples import Echo, FastTrace

from typing import Any

import copy
import numpy as np

class EchoesWithPulses(VirtualDatasource):
    """Echoes, with a number of data points from the waveforms corresponding to the pulses that resulted in each echo."""

    def __init__(self, reference_sensor:str, dependencies:list, pulse_sample_size:int=10, zero_baseline:bool=True, decimation:int=1, binning:int=1, pulses_as_echoes:bool=True):
        """Constructor
            Args:
                reference_sensor (str): The name of the sensor (e.g. 'pixell_bfc').
                dependencies (list): A list of the datasource names. 
                    First element should be an Echo datasource (e.g. 'pixell_bfc_ech')
                    Second element should be a Trace datasource (e.g. 'pixell_bfc_ftrr')
                pulse_sample_size (int): The number of data points to gather before and after each echo in the waveforms.
                    For example, with pulse_sample_size=10, the pulses will be 21 points, because the highest point is
                    taken in addition to the 10 points before and the 10 points after.
                zero_baseline (bool): If True, the baseline of the waveforms are re-calibrated to be around zero.
                decimation (int): Must be an integer larger than 1. Downsample the pulses by this factor by decimation.
                binning (int): Must be an integer larger than 1. Downsample the pulses by this factor by binning.
                pulses_as_echoes (bool): If True, each sample from each pulse is treated as an echo. 
                    If False, the echoes are untouched and the pulses are stored under the 'pulses' key in the raw dictionnary. 
        """
        super(EchoesWithPulses, self).__init__(f'ech-pulses', dependencies, None)
        self.reference_sensor = reference_sensor
        self.original_echoes_datasource = dependencies[0]
        self.original_trace_datasource = dependencies[1]
        self.pulse_sample_size = pulse_sample_size
        self.pulses_as_echoes = pulses_as_echoes

        trace_processing_list = []
        if zero_baseline:
            trace_processing_list.append(ZeroBaseline())
        if binning != 1:
            trace_processing_list.append(Binning(binning))
        if decimation != 1:
            trace_processing_list.append(Decimate(decimation))
        self.trace_processing = TraceProcessingCollection(trace_processing_list)  

    def __getitem__(self, key:Any):

        if isinstance(key, slice):
            return self[platform.slice_to_range(key, len(self))]
        if isinstance(key, range):
            return [self[index] for index in key]

        echoes_sample = self.datasources[self.original_echoes_datasource][key]

        timestamp = echoes_sample.timestamp
        specs = echoes_sample.specs

        pulses, distance_scaling = echoes_sample.get_pulses(self.original_trace_datasource.split('_')[-1], self.pulse_sample_size, self.trace_processing, return_distance_scaling=True)

        if not self.pulses_as_echoes:

            raw = clouds.to_echo_package(
                indices = echoes_sample.indices, 
                distances = echoes_sample.distances, 
                amplitudes = echoes_sample.amplitudes,
                timestamps = echoes_sample.timestamps, 
                flags = echoes_sample.flags, 
                timestamp = timestamp,
                specs = {"v" : specs['v'], "h" : specs['h'], "v_fov" : specs['v_fov'], "h_fov" : specs['h_fov']},
                distance_scale = 1.0, 
                amplitude_scale = 1.0, 
                led_power = 1.0, 
                eof_timestamp = None)

            raw['pulses'] = pulses
            
            return Echo(key, self, raw, timestamp)

        else:

            indices = np.repeat(echoes_sample.indices[...,None], pulses.shape[-1], axis=-1).flatten()
            delta = self.pulse_sample_size*distance_scaling
            deltas = np.repeat(np.linspace(-delta, delta, pulses.shape[-1])[None,...], pulses.shape[0], axis=0)
            distances = (np.repeat(echoes_sample.distances[...,None], pulses.shape[-1], axis=-1) - deltas).flatten()
            amplitudes = pulses.flatten()
            timestamps = (np.repeat(echoes_sample.timestamps[...,None], pulses.shape[-1], axis=-1)).flatten()
            flags = (np.repeat(echoes_sample.flags[...,None], pulses.shape[-1], axis=-1)).flatten()

            keep = np.where((distances > 0) & (amplitudes > 0))[0]

            raw = clouds.to_echo_package(
                indices = indices[keep], 
                distances = distances[keep], 
                amplitudes = amplitudes[keep],
                timestamps = timestamps[keep], 
                flags = flags[keep], 
                timestamp = timestamp,
                specs = {"v" : specs['v'], "h" : specs['h'], "v_fov" : specs['v_fov'], "h_fov" : specs['h_fov']},
                distance_scale = 1.0, 
                amplitude_scale = 1.0, 
                led_power = 1.0, 
                eof_timestamp = None)
        
            return Echo(key, self, raw, timestamp)


from pioneer.das.api.samples.point_cloud import PointCloud

import numpy as np

class XYZIT(PointCloud):
    """Point cloud provided by a mechanical lidar sensor.
        For each data point, contains (x,y,z) coordinates, the intensity (i) and a timestamp (t).
    """

    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super().__init__(index, datasource, virtual_raw, virtual_ts)
    
    def get_point_cloud(self, referential=None, ignore_orientation=False, undistort=False, reference_ts=-1, dtype=np.float64):
        """Compute a 3D point cloud from raw data
        
        Args:
            referential: The target sensor referential or full datasource name
            ignore_orientation: Ignore the source sensor orientation (default: {False})
            undistort: Apply motion compensation to 3d points.
            reference_ts:  (only used if referential == 'world' and/or undistort == True), 
                           refer to compute_transform()'s documentation
            dtype: the output numpy data type
        """
        pts_Local = np.concatenate([self.get_field('x').reshape(-1, 1),
                            self.get_field('y').reshape(-1, 1),
                            self.get_field('z').reshape(-1, 1),], axis=1).astype(dtype)

        # Legacy stuff with Ouster64's temperature compensation.
        if hasattr(self.datasource.sensor, "get_corrected_cloud"):
            pts_Local = self.datasource.sensor.get_corrected_cloud(self.timestamp, pts_Local, dtype)

        if undistort:
            to_world = referential == 'world'
            self.undistort_points(pts_Local, self.get_field('t'), reference_ts, to_world, dtype = dtype)
            if to_world:
                return pts_Local # note that in that case, orientation has to be ignored
        
        return self.transform(pts_Local, referential, ignore_orientation, reference_ts, dtype = dtype)


from pioneer.common import misc, linalg
from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.datasources import DatasourceWrapper, VirtualDatasource
from pioneer.das.api.interpolators import nearest_interpolator
from pioneer.das.api.samples import Sample
from pioneer.das.api.samples import annotations

from typing import Optional, List, Dict, Tuple, Any

import glob
import numpy as np
import os
import pickle

class Sensor(object):
    """A sensor encapsulate a sensor and its datasources. 
    *Important* when you add a new derivation of class Sensor, 
    don't forget to add it to platform.SENSOR_FACTORY.
    """
    class NoPathToReferential(Exception):
        pass

    def __init__(self, name:str, pf:'Platform', factories:Dict[str, Tuple[Any, Any]] = {}):
        """Constructor.
        
        Args:
            name: the sensor's name and position id, e.g. 'eagle_tfc'
            platform: the platform this sensor belongs to
            factories: a dict containing one entry per datasource type, each entry's value is a tuple containing 
                Sample-derived, optional_interpolator_function. For example, a factory dict could be 
                {"ech":(Echo, None), "sta":(Sample, interpolators.linear_dict_of_float_interpolator)}
        """
        self.name = name
        self.datasources = {}
        self.factories = {**factories, **annotations.ANNOTATIONS_FACTORY}
        self.pf = pf
        self.yml = {}
        try:
            self.yml = pf.yml[name]
        except:
            pass
        self.orientation = None
        self.extrinsics = {}
        self.intrinsics = None
        self.egomotion_provider = None
        self._extrinsics_dirty = misc.Signal()
        self.pcl_datasource = None

    def datasource_names(self) -> List[str]:
        """Returns this sensor's datasource types"""
        return [ds.label for ds in self.datasources.values()]

    @property
    def platform(self):
        return self.pf

    @property
    def extrinsics_dirty(self):
        """ The extrinsics dirty signal """
        return self._extrinsics_dirty

    def invalidate_datasources_caches(self):
        """ Invalidate sensor's datasources caches """
        for ds in self.datasources.values():
            ds.invalidate_caches()
            
    def add_datasource(self, ds, ds_type:str, cache_size:int=100):
        """Adds a datasource to this sensor

        Args:
            ds: a Filesource-derived instance or VirtualDatasource
            ds_type: the datasource type, e.g. 'ech'
        """

        if isinstance(ds, VirtualDatasource):
            ds._set_sensor(self)
            self.datasources[ds_type] = ds
        
        else:
            ds_type_no_suffix = ds_type.split('-')[0]
            dsw = DatasourceWrapper(self, ds_type, ds, self.factories[ds_type_no_suffix] if ds_type_no_suffix in self.factories else (Sample, nearest_interpolator), cache_size=cache_size)
            self.datasources[ds_type] = dsw

    def load_intrinsics(self, intrinsics_config:str):
        """Looks for a pickle file containing intrinsics information for this sensor, e.g. 'eagle_tfc.pkl'

        Args:
            intrinsics_config: path to folder containing this sensor's intrinsics pickle file, 
            (absolute or relative to dataset path), e.g. '/nas/cam_intrinsics' or 'cam_intrinsics'
        """
        
        paths = glob.glob(os.path.join(self.pf.try_absolute_or_relative(intrinsics_config), '{}*results.pkl'.format(self.name)))
        if paths:
            path = paths[0]
            if len(paths) > 1:
                LoggingManager.instance().warning('more than one intrinsics, using {}'.format(path))
            with open(path, 'rb') as f:
                self.intrinsics = pickle.load(f)

    def load_extrinsics(self, extrinsics_folder:str):
        """Looks for a pickle file containing extrinsics information for this sensor, named 'From-To' e.g. 'flir_tfl-eagle_tfc.pkl'

        Args:
            intrinsics_config: path to folder containing this sensor's extrinsics pickle file 
            (absolute or relative to dataset path), e.g. '/nas/extrinsics' or 'extrinsics'
        """
        targets = {}

        for target in self.pf.yml.keys():
            if self.name == target:
                targets[target] = np.eye(4, dtype = 'f8')
                continue

            # try to find self.name -> target mapping
            extrinsics_folder_path = self.pf.try_absolute_or_relative(extrinsics_folder)

            paths = glob.glob(os.path.join(extrinsics_folder_path, f"{self.name}-{target}.pkl"))

            if paths:
                with open(paths[0], 'rb') as f:
                    targets[target] = pickle.load(f).astype('f8')
            else:
                # try to find target -> self.name mapping instead
                paths = glob.glob(os.path.join(extrinsics_folder_path, f"{target}-{self.name}.pkl"))
                if paths:
                    with open(paths[0], 'rb') as f:
                        targets[target] = linalg.tf_inv(pickle.load(f)).astype('f8')


        self.extrinsics = targets

    def create_egomotion_provider(self) -> Optional['EgomotionProvider']:
        return

    def map_to(self, target:str) -> np.ndarray:
        """Returns a 4x4 tranform matrix mapping a point from this sensor's referential to 'target' sensor's referential
        
        Args:
            target: the name of the target sensor in which referential we want to map to
        Raises:
            Sensor.NoPathToReferential: if no mapping could be found
        """

        return self._map_to_recurse(target, self.name, set([self.name]))

    def _map_to_recurse(self, target, orig, visited):

        if target in self.extrinsics:
            return self.extrinsics[target]

        for alt_target,v in self.extrinsics.items():
            if alt_target in visited:
                continue
            try:
                visited.update([alt_target])
                m = self.pf.sensors[alt_target]._map_to_recurse(target, alt_target, visited)
                return np.matmul(m, v)
            except:
                continue # we will try another path...

        raise Exception()


    def keys(self):
        """Returns the datasource types, implement dict API"""
        return self.datasources.keys()

    def items(self):
        """Returns the datasource key,value iterable, implement dict API"""
        return self.datasources.items()

    def __contains__(self, key:str):
        return key in self.datasources

    def __len__(self):
        """Returns the number of datasources in this sensor."""
        return len(self.datasources)

    def __getitem__(self, key:str):
        try: return self.datasources[key]
        except KeyError:
            raise KeyError('This data source for sensor {} does not exist.'.format(self.name))



from pioneer.common                                import linalg
from pioneer.das.api.egomotion.egomotion_provider  import EgomotionProvider
from pioneer.das.api.datasources                   import DatasourceWrapper

import numpy as np



class CarlaEgomotionProvider(EgomotionProvider):
    """Ego Motion provider for simulated datasets (CARLA simulator)"""

    def __init__(self, referential_name:str, carla_imu_datasource:DatasourceWrapper):
        super().__init__(referential_name)
        self.carla_imu_datasource = carla_imu_datasource

    def get_transform(self, timestamp:int) -> np.ndarray:
        raw = self.carla_imu_datasource.get_at_timestamp(timestamp).raw
        if type(raw) is dict: raw = raw['data']
        return linalg.tf_from_poseENU([raw['x'],raw['y'],raw['z'],raw['roll'],raw['pitch'],-raw['yaw']])

    def get_timestamps(self) -> np.ndarray:
        return self.carla_imu_datasource.timestamps



from pioneer.common import linalg
from typing import Optional
from abc import ABC, abstractclassmethod

import numpy as np
import warnings
warnings.simplefilter('once', DeprecationWarning)



class EgomotionProvider(ABC):
    """Computes the position and orientation of the ego vehicle. A Platform must have
        an ego motion provider in order to use the 'world' referential.
    """

    def __init__(self, referential_name:str):
        self.referential_name = referential_name

    @abstractclassmethod
    def get_transform(self, timestamp:int) -> np.ndarray:
        """Returns the transform to the world from the Egomotion Provider's referential at a given time."""

    @abstractclassmethod
    def get_timestamps(self) -> np.ndarray:
        """Returns all timestamps in an array."""

    def get_inverse_transform(self, timestamp:int) -> np.ndarray:
        """Returns the transform to the Egomotion Provider's referential from the world at a given time."""
        return linalg.tf_inv(self.get_transform(timestamp))

    def get_first_transform(self) -> np.ndarray:
        """Returns the transform at the first available timestamp"""
        return self.get_transform(self.get_timestamps()[0])

    def get_first_inverse_transform(self) -> np.ndarray:
        """Returns the inverse transform at the first available timestamp"""
        return self.get_inverse_transform(self.get_timestamps()[0])

    def get_trajectory(self, timestamps:Optional[np.ndarray]=None, start_at_origin:bool=False, subsampling:int=100) -> np.ndarray:
        """Returns all transforms."""
        timestamps = self.get_timestamps() if timestamps is None else timestamps
        end = int(np.ceil(timestamps.shape[0]/subsampling))
        trajectory = np.empty((end,4,4))
        first_inverse_transform = self.get_first_inverse_transform() if not start_at_origin else np.eye(4)
        for i in range(end):
            to_world = self.get_transform(timestamps[i*subsampling])
            trajectory[i] = first_inverse_transform @ to_world
        return trajectory






    ### Legacy section ###

    def get_Global_from_Ego_at(self, ts:int, dtype = np.float64) -> np.ndarray:
        warnings.warn("EgomotionProvider.get_Global_from_Ego_at() is deprecated. Use EgomotionProvider.get_transform() instead.", DeprecationWarning)
        return self.get_transform(timestamp = ts).astype(dtype)

    def get_timestamps_range(self) -> np.ndarray:
        warnings.warn("EgomotionProvider.get_timestamps_range() is deprecated. Use EgomotionProvider.get_timestamps() instead (first and last elements).", DeprecationWarning)
        _ts = self.get_timestamps()
        return np.array((_ts[0], _ts[-1]))

    @property 
    def tf_Global_from_EgoZero(self) -> np.ndarray:
        warnings.warn("EgomotionProvider.tf_Global_from_EgoZero is deprecated. Use EgomotionProvider.get_first_transform() instead.", DeprecationWarning)
        return self.get_first_transform()

    def compute_tf_EgoZero_from_Sensor(self, tf_Ego_from_Sensor:np.ndarray, reference_ts:int) -> np.ndarray:
        warnings.warn("EgomotionProvider.compute_tf_EgoZero_from_Sensor(tf_Ego_from_Sensor, reference_ts) is deprecated. \
            Use: EgomotionProvider.get_first_inverse_transform() @ EgomotionProvider.get_transform(reference_ts) @ tf_Ego_from_Sensor ", 
        DeprecationWarning)
        return self.get_first_inverse_transform() @ self.get_transform(reference_ts) @ tf_Ego_from_Sensor

    def compute_trajectory(self, sorted_ts:np.ndarray, tf_Global_from_EgoZero:np.ndarray=np.eye(4), dtype = np.float64, subsampling:int=100) -> np.ndarray:
        warnings.warn("EgomotionProvider.compute_trajectory() is deprecated. Use EgomotionProvider.get_trajectory() instead.", DeprecationWarning)
        return self.get_trajectory(sorted_ts, subsampling=subsampling)

    def apply_trajectory(self, trajectory:np.ndarray, points:np.ndarray, subsampling:int=100) -> np.ndarray:
        warnings.warn("EgomotionProvider.apply_trajectory() is deprecated.", DeprecationWarning)
        corrected = np.empty_like(points)
        n_points = points.shape[0]
        for i in range(trajectory.shape[0]):
            s, e = i * subsampling, min((i+1) * subsampling, n_points)
            corrected[s:e] = linalg.map_points(trajectory[i], points[s:e])
        return corrected



from pioneer.common import banks, clouds, constants
from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.sensors.lcax import LCAx

from typing import Union

import numpy as np


class Pixell(LCAx):

    def __init__(self, name: str, platform: 'Platform'):
        super(Pixell, self).__init__(name, platform)
        self.orientation = np.array([[ 0, 0, 1],
                                     [-1, 0, 0],
                                     [ 0, 1, 0]], 'f8')
        self.adc_freq = 100e6
        self.specs = {'v':8, 'h':96, 'v_fov':16.073, 'h_fov':180.604}
        self.head_positions = constants.PIXELL_HEAD_POSITIONS
        self.head_position_per_channel = self.get_head_position_per_channel()
        self.calibrated_angles = None

    def load_intrinsics(self, intrinsics_config: Union[str, dict]):
        '''Basic intrinsics are brougth back under Sensor.
        '''
        super(Pixell, self).load_intrinsics(intrinsics_config)
        try:
            cfg = self['cfg'][0].raw
            self.azimut = np.deg2rad(cfg['calibration']['ID_CHANNEL_ANGLE_AZIMUT'])
            self.elevation = np.deg2rad(cfg['calibration']['ID_CHANNEL_ANGLE_ELEVATION'])
            self.calibrated_angles = np.c_[self.elevation, self.azimut].astype(float)
            _time_base_delays = cfg['calibration']['ID_TIMEBASE_DELAY']
            self.time_base_delays = {
                'low': _time_base_delays[:self.specs['v']*self.specs['h']][constants.ID_FROM_LCAS_CHAN_TO_SENSOR_CHAN],
                'high':_time_base_delays[self.specs['v']*self.specs['h']:][constants.ID_FROM_LCAS_CHAN_TO_SENSOR_CHAN],
            }
        except:
            pass

    def cache(self, specs):
        hashable = frozenset([(k,specs[k]) for k in ['v', 'h', 'v_fov', 'h_fov']])
        if not hashable in self._cache:
            cache = super(Pixell, self).cache(specs)
            cache['lcas_offsets'] = self.head_position_per_channel
            if self.calibrated_angles is not None:
                cache['angles'] = self.calibrated_angles
                cache['quad_angles'] = clouds.custom_quad_angles(specs, cache['angles'])
            else:
                cache['quad_angles'] = clouds.quad_angles(specs)
            self._cache[hashable] = cache

        return self._cache[hashable]

    def get_trace_smoothing_kernel(self):
        return constants.PIXELL_FILTERS_KERNEL

    def get_corrected_cloud(self, timestamp, cache, type, indices, distances, amplitudes=None, dtype=np.float64):
        
        if self.calibrated_angles is not None:
            type_pts, type_amplitudes, type_indices = self.get_cloud(cache, type, indices, distances, amplitudes, dtype)

            if type == 'point_cloud':
                return type_pts
            elif type == 'quad_cloud':
                return type_pts, type_amplitudes, type_indices
            else:
                raise KeyError('Can not compute the point cloud projection')
        else:
            return super().get_corrected_cloud(timestamp, cache, type, indices, distances, amplitudes, dtype)

    def get_cloud(self, cache, type, indices, distances, amplitudes=None, dtype=np.float64):
        '''Get cloud data (type_pts, type_amplitudes, type_indices) from a sensor.'''

        if type == 'point_cloud':
            x,y,z = self.xyz_from_heads(distances, cache['angles'][indices], cache['lcas_offsets'][indices])
            pts = np.stack(( x,  y, - z), axis=-1)

            pts = pts @ self.orientation
            return pts, amplitudes, indices
        
        elif type == 'quad_cloud':
            distances_, elevation, azimut = [],[],[]
            m_ = cache['corrected_specs']['v'] * cache['corrected_specs']['h']
            for i in range(4):
                elevation_, azimut_ = self.angles_from_heads(distances, cache['quad_angles'][indices + i * m_], cache['lcas_offsets'][indices])
                distances_.extend(distances)
                elevation.extend(elevation_)
                azimut.extend(azimut_)

            distances_ = np.array(distances_, dtype=float)
            elevation = np.array(elevation, dtype=float)
            azimut = np.array(azimut, dtype=float)
            y,z,x = clouds.direction_spherical(elevation, azimut)
            pts = np.stack((distances_ * x, distances_ * y, -distances_ * z), axis=-1)
            pts = pts @ self.orientation
            quad_amplitudes = clouds.quad_stack(amplitudes)
            quad_indices = clouds.generate_quads_indices(indices.shape[0], np.uint32).flatten()

            return pts, quad_amplitudes, quad_indices

    @staticmethod
    def angles_from_heads(distances, angles, head_positions):
        distances_ = distances-np.cos(angles[:,0]) * (head_positions[:,1] * np.sin(angles[:,1]) + head_positions[:,0] * np.cos(angles[:,1])) 
        distances_ -= head_positions[:,2] * np.sin(angles[:,0])

        l = distances_ * np.cos(angles[:,0])
        x,y,z = (l * np.cos(angles[:,1]), l * np.sin(angles[:,1]), distances_ * np.sin(angles[:,0]))
        
        x += head_positions[:,0]
        y += head_positions[:,1]  
      
        return np.arctan2(z, (x**2 + y**2)**0.5), np.arctan2(y, x)

    @staticmethod
    def xyz_from_heads(distances, angles, head_positions):
       
        distances_ = distances-np.cos(angles[:,0]) * (head_positions[:,1] * np.sin(angles[:,1]) + head_positions[:,0] * np.cos(angles[:,1])) 
        distances_ -= head_positions[:,2] * np.sin(angles[:,0])

        l = distances_ * np.cos(angles[:,0])
        x,y,z = (l * np.cos(angles[:,1]), l * np.sin(angles[:,1]), distances_ * np.sin(angles[:,0]))
        
        x += head_positions[:,0]
        y += head_positions[:,1]  
       
        return x,y,z

    def get_head_position_per_channel(self):
        head = np.zeros((3, self.specs['v'], self.specs['h']), dtype=float)
        head[:,:,:32] = self.head_positions['left'][:,None,None]
        head[:,:,32:64] = self.head_positions['center'][:,None,None]
        head[:,:,64:] = self.head_positions['right'][:,None,None]
        return np.c_[head[0].ravel(),head[1].ravel(),head[2].ravel()]


from pioneer.das.api.egomotion.egomotion_provider import EgomotionProvider
from pioneer.das.api.egomotion.carla_egomotion_provider import CarlaEgomotionProvider
from pioneer.das.api.egomotion.imu_egomotion_provider import IMUEgomotionProvider
from pioneer.das.api.egomotion.vaya_egomotion_provider import VayaEgomotionProvider


from pioneer.das.api.sensors.lcax import LCAx

import numpy as np
        
class LCA3(LCAx):
    def __init__(self, name: str, platform: 'Platform'):
        super(LCA3, self).__init__(name, platform)
        self.scan_direction = 'horizontal'
        #90 deg around 'z' axis
        self.orientation = np.array([[ 0,  1,  0],
                                        [-1,  0,  0],
                                        [ 0,  0,  1]], 'f8')

    def get_trace_smoothing_kernel(self):
        """Get the convolution kernel for trace processing for Eagle"""
        # Temporary fix for binning while we can't read the binning parameter anywhere
        kernel = np.array([1,9,18,23,26,23,8,-7,-14,-17,-17,-15,-13,-11,-10,-8,-8,-6,-5,-4,-3,-3,-3,-2,-2])[::-1]/256
        if self.binning == 2:
            kernel = np.array([1,18,26,8,-14,-17,-13,-10,-8,-5,-3,-3,-2])[::-1]/128
        return kernel



from pioneer.common import platform
from pioneer.das.api import categories, datatypes
from pioneer.das.api.samples.sample import Sample

import copy
import cv2
import numpy as np


class Seg2d(Sample):

    def __init__(self, index, datasource, virtual_raw=None, virtual_ts=None):
        super(Seg2d, self).__init__(index, datasource, virtual_raw, virtual_ts)

    def poly2d(self, confidence_threshold=0.5):

        raw = copy.deepcopy(self.raw)

        poly2d = np.empty(0, dtype=datatypes.poly2d())

        box_confidences = None
        if 'box_confidence' in raw.keys():
            box_confidences = raw['box_confidence']

        for i, seg_data in enumerate(raw['data']):
            if box_confidences is not None and box_confidences[i] < confidence_threshold:
                continue
            mask = (seg_data['confidences'] > confidence_threshold).astype(np.uint8)
            if mask.max() == 0:
                continue
            polys, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            polys = np.array(polys)
            for poly in polys:
                poly = poly[:, 0, :]
                poly2d = np.append(poly2d, np.array([(poly, seg_data['classes'], 0, 0)], dtype=datatypes.poly2d()))

        return poly2d

    def mask_category(self, category: str, resolution: tuple = None, confidence_threshold: float = 0.5):

        _, _, ds_type = platform.parse_datasource_name(self.datasource.label)
        source = categories.get_source(ds_type)

        mask = np.zeros((self.raw['data']['confidences'][0].shape))

        for seg_data in self.raw['data']:
            name, _ = categories.get_name_color(source, seg_data['classes'])
            if name == category:
                above_threshold = np.where(seg_data['confidences'] >= confidence_threshold)
                mask[above_threshold] = 1
                break

        return self.resize_mask(mask, resolution) if resolution is not None else mask

    @staticmethod
    def resize_mask(mask, resolution):
        return cv2.resize(mask, (resolution[1], resolution[0]), interpolation=cv2.INTER_NEAREST)



from pioneer.das.api.samples.image import Image

import numpy as np


class ImageCylinder(Image):
    """Image derivation that was transformed by a cylindrical projection"""
    
    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super().__init__(index, datasource, virtual_raw, virtual_ts)
        self.cylindrical_projection = datasource.cylindrical_projection

    def project_pts(self, pts, mask_fov=False, output_mask=False, undistorted=False, margin=0):
        ''' project 3D in the 2D cylindrical referiencial

            Args:
                pts: 3D point in the center camera referential (3xN)
                mask_fov (optionnal): removes points outside the fov
                output_mask (optionnal): if True, returns the mask applied to the points
                undistorted (=False): Does nothing in the case of ImageCylinder, as it is always undistorted
                margin (optionnal): margin (in pixels) outside the image unaffected by the fov mask
            Return:
                2xN: 2D points in cylindrical image referential
                mask (optionnal): returned if output_mask is True
        '''
        return self.cylindrical_projection.project_pts(pts.T, mask_fov, output_mask, margin)

    def get_image(self, undistort: bool = False) -> np.ndarray:
        return super().get_image(undistort=False) # ImageCylinder is always undistorted
        




from pioneer.das.api.samples.sample import Sample

from typing import Optional, Tuple

import numpy as np
import warnings
warnings.simplefilter('once', DeprecationWarning)


class PointCloud(Sample):

    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super().__init__(index, datasource, virtual_raw, virtual_ts)

    def get_point_cloud(self, referential=None, ignore_orientation=False, undistort=False, reference_ts=-1, dtype=np.float64) -> np.ndarray:
        """Compute a 3D point cloud from raw data
        
        Args:
            referential: The target sensor referential or full datasource name
            ignore_orientation: Ignore the source sensor orientation (default: {False})
            undistort: Apply motion compensation to 3d points.
            reference_ts:  (only used if referential == 'world' and/or undistort == True), 
                           refer to compute_transform()'s documentation
            dtype: the output numpy data type
        """

        points = np.concatenate([self.get_field('x').reshape(-1, 1),
                                 self.get_field('y').reshape(-1, 1),
                                 self.get_field('z').reshape(-1, 1),
            ], axis=1).astype(dtype)

        if undistort:
            to_world = referential == 'world'
            self.undistort_points(points, self.get_field('t'), reference_ts, to_world, dtype = dtype)
            if to_world:
                return points # note that in that case, orientation has to be ignored
        
        return self.transform(points, referential, ignore_orientation, reference_ts, dtype = dtype)

    @property
    def fields(self) -> Tuple[str]:
        return self.raw.dtype.names

    def get_field(self, field:str) -> Optional[np.ndarray]:
        if field in self.fields: 
            return self.raw[field]

    @property
    def distances(self):
        point_cloud = self.get_point_cloud()
        return (point_cloud[:,0]**2 + point_cloud[:,1]**2 + point_cloud[:,2]**2)**0.5

    @property
    def size(self) -> int:
        return self.raw.shape[0]



    ### Legacy section ###
    
    @property
    def timestamps(self):
        warnings.warn("PointCloud.timestamps is deprecated. Use PointCloud.get_field('t') instead.", DeprecationWarning)
        return self.get_field('t')
    
    @property
    def amplitudes(self):
        warnings.warn("PointCloud.amplitudes is deprecated. Use PointCloud.get_field('i') instead.", DeprecationWarning)
        return self.get_field('i').astype('f4')

    def get_cloud(self, referential=None, ignore_orientation=False, undistort=False, reference_ts=-1, dtype=np.float64):
        warnings.warn("PointCloud.get_cloud() is deprecated. Use (PointCloud.get_point_cloud(), PointCloud.get_field('i'), np.arange(PointCloud.size)) instead.", DeprecationWarning)
        points = self.get_point_cloud(referential, ignore_orientation, undistort, reference_ts, dtype)
        return points, self.get_field('i'), np.arange(self.size)

    def point_cloud(self, referential=None, ignore_orientation=False, undistort=False, reference_ts=-1, dtype=np.float64):
        warnings.warn("PointCloud.point_cloud() is deprecated. Use PointCloud.get_point_cloud() instead.", DeprecationWarning)
        return self.get_point_cloud(referential, ignore_orientation, undistort, reference_ts, dtype)
    
    def get_rgb_from_camera_projection(pcloud_ds:Sample, camera_ds:str, undistort:bool=False, return_mask:bool=False):
        from pioneer.das.api.datasources.virtual_datasources.rgb_cloud import get_rgb_from_camera_projection
        warnings.warn("PointCloud.get_rgb_from_camera_projection() is deprecated. Use from pioneer.das.api.datasources.virtual_datasources.rgb_cloud import get_rgb_from_camera_projection instead.", DeprecationWarning)
        return get_rgb_from_camera_projection(pcloud_ds, camera_ds, undistort, return_mask)


from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.sensors.lidar  import Lidar
from pioneer.das.api.samples        import Sample, XYZIT
from pioneer.das.api.interpolators  import nearest_interpolator

from typing import Any, Dict, Tuple, Union

import numpy as np

class MotorLidar(Lidar):
    
    def __init__(self, name: str, platform: 'Platform', factories:Dict[str, Tuple[Any, Any]] = {}):
        super().__init__(name, platform, 
            factories = {**factories,
                'xyzit': (XYZIT, nearest_interpolator), 
                'temp': (Sample, nearest_interpolator),
            })
        
        self.temperature_slope = None
        self.temperature_reference = None
        self.temperature_coeffs = None
        self.temperature_compensation = False
        
        self.pcl_datasource = 'xyzit'

    def load_intrinsics(self, intrinsics_config: Union[str, dict]):
        '''load data from yml platform '''
        temperature_config = intrinsics_config.get('temperature')
        if temperature_config:
            self.temperature_compensation = True
            self.temperature_slope = temperature_config['slope']
            self.temperature_reference = temperature_config['reference']
            self.temperature_coeffs = np.array([self.temperature_slope, 
                                            self.temperature_reference])
        
        orientation = intrinsics_config.get('orientation')
        if orientation:
            self.orientation = np.array(orientation)

    def apply_temperature_correction(self, timestamp, distances):
        """Applies temperature-related distance corrections"""
        if (self.temperature_slope is None and
            self.temperature_reference is None):
            self.temperature_compensation = False
            LoggingManager.instance().warning('MotorLidar Log: Temperature compensation mode is deactivated')
            return distances
        
        offsets_ = 0
        try:
            temperature = self.datasources['temp'].get_at_timestamp(timestamp).raw['data']
            k = len(self.temperature_coeffs)-1
            offsets_ -= np.sum([temperature**(k-i) * self.temperature_coeffs[i] for i in range(k)])
        except Exception as e:
            temperature = None
            LoggingManager.instance().warning('Failed to apply the temperature for the distance correction. '
                                'Error: {}'.format(str(e)))
        offsets_ -= self.temperature_coeffs[-1]

        #print(f'Log: 3D points for {self.name} are corrected with temperature T={temperature}: global offset is {offsets_}')
        return distances + offsets_
        
    def get_corrected_cloud(self, timestamp, pts, dtype):
        '''It corrects the pts-cloud according to a temperature compensation (if any).'''
        if not self.temperature_compensation:
            return pts
        distances = np.linalg.norm(pts, axis=1, keepdims=True)
        directions = pts / distances
        distances = self.apply_temperature_correction(timestamp, distances)
        pts = directions * distances
        return pts



from pioneer.das.api.interpolators   import linear_ndarray_interpolator
from pioneer.das.api.samples import Sample
from pioneer.das.api.sensors.sensor import Sensor
from pioneer.das.api.egomotion.vaya_egomotion_provider import VayaEgomotionProvider



class GPSVaya(Sensor):
    def __init__(self, name: str, platform: 'Platform'):
        factories = {'mat':(Sample, linear_ndarray_interpolator)}
        super().__init__(name, platform, factories)

    def create_egomotion_provider(self):
        self.egomotion_provider = VayaEgomotionProvider(self.name, self["mat"])
        return self.egomotion_provider



from pioneer.common import platform as platform_utils
from pioneer.common.cylindrical_projection import CylindricalProjection, Pos
from pioneer.das.api.datasources.virtual_datasources import VirtualDatasource
from pioneer.das.api.interpolators import nearest_interpolator
from pioneer.das.api.samples import ImageCylinder

import numpy as np

class FlirCylindricalProjection(VirtualDatasource):
    """Merges the images of three cameras in a single one."""

    def __init__(self, reference_sensor:str, dependencies:list, 
                    radius=50.0, fov_h=np.deg2rad(210), fov_v=np.deg2rad(67.5), 
                    image_h=2000, image_v=int(2000*0.25), 
                    fusion_overlap_ratio=0.25):
        """Constructor
            Args:
                reference_sensor (str): The name of the central camera (e.g. 'flir_bfc').
                dependencies (list): A list of the image datasource names from left to right. 
                    (e.g. ['flir_bfl_img','flir_bfc_img','flir_bfr_img'])
        """
        super(FlirCylindricalProjection, self).__init__('img-cyl', dependencies, None, timestamps_source = dependencies[Pos.CENTER.value])
        self.reference_sensor = reference_sensor
        self.dependencies = dependencies
        self.config={'radius':radius, 'FOV_h':np.deg2rad(fov_h), 'FOV_v':np.deg2rad(fov_v), 'image_h':image_h, 'image_v':int(image_v), 'fusion_overlap_ratio': fusion_overlap_ratio}
        self.cylindrical_projection = None
            

    def _set_cylindrical_projection(self):
        pf = self.datasources[self.dependencies[0]].sensor.platform
        cameras = list(map(lambda camera: platform_utils.extract_sensor_id(camera), self.dependencies))
        intrinsics_calibrations = list(map(lambda camera: pf.intrinsics[camera]['matrix'], cameras))
        distortion_coefficients = list(map(lambda camera: pf.intrinsics[camera]['distortion'], cameras))
        extrinsic_calibrations  = list(map(lambda camera: pf.extrinsics[camera][cameras[Pos.CENTER.value]], cameras))
            
        self.cylindrical_projection = CylindricalProjection(intrinsics_calibrations, distortion_coefficients, extrinsic_calibrations, self.config)


    def __getitem__(self, key):

        if self.cylindrical_projection is None:
            self._set_cylindrical_projection()

        img_c_sample = self.datasources[self.dependencies[Pos.CENTER.value]][key]
        ts = img_c_sample.timestamp

        img_c = img_c_sample.raw
        img_r = self.datasources[self.dependencies[Pos.RIGHT.value]].get_at_timestamp(ts, nearest_interpolator).raw
        img_l = self.datasources[self.dependencies[Pos.LEFT.value]].get_at_timestamp(ts, nearest_interpolator).raw

        raw = self.cylindrical_projection.stitch([img_l, img_c, img_r])

        return ImageCylinder(key, self, raw, virtual_ts=ts)




from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource
from pioneer.das.api.datatypes import datasource_xyzit
from pioneer.das.api.samples.point_cloud import PointCloud

from typing import Any

import numpy as np


class PointCloudFusion(VirtualDatasource):
    """Merges multiple points clouds in a single one"""

    def __init__(self, reference_sensor:str, dependencies:list):
        """Constructor
            Args:
                reference_sensor (str): The name of the reference sensor (e.g. 'pixell_bfc').
                dependencies (list): A list of the point cloud datasources to fuse. 
        """
        ds_type = 'xyzit-fused'
        super(PointCloudFusion, self).__init__(ds_type, dependencies, None)
        self.reference_sensor = reference_sensor  
        self.reference_datasource = dependencies[0]  

    def get_at_timestamp(self, timestamp):
        sample = self.datasources[self.reference_datasource].get_at_timestamp(timestamp)
        return self[int(np.round(sample.index))]

    @staticmethod
    def stack_point_cloud(stack, point_cloud):
        return np.vstack([stack, point_cloud])

    def clear_cache(self):
        self.local_cache = None

    def __getitem__(self, key:Any):

        pcloud_fused = np.empty((0,5))

        for ds_name in self.dependencies:
            sample = self.datasources[ds_name][key]
            xyz = sample.get_point_cloud(referential=self.reference_sensor)

            if self.sensor.orientation is not None:
                xyz = xyz @ self.sensor.orientation

            pcloud = np.empty((xyz.shape[0],5))
            pcloud[:,[0,1,2]] = xyz
            pcloud[:,3] = sample.get_field('i')
            pcloud[:,4] = sample.get_field('t')

            pcloud_fused = self.stack_point_cloud(pcloud_fused, pcloud)

        #package in das format
        dtype = datasource_xyzit()
        raw = np.empty((pcloud_fused.shape[0]), dtype=dtype)
        raw['x'] = pcloud_fused[:,0]
        raw['y'] = pcloud_fused[:,1]
        raw['z'] = pcloud_fused[:,2]
        raw['i'] = pcloud_fused[:,3]
        raw['t'] = pcloud_fused[:,4]

        ts = self.datasources[self.dependencies[0]][key].timestamp
        return PointCloud(index=key, datasource=self, virtual_raw=raw, virtual_ts=ts)




from pioneer.das.api.samples.sample import Sample

from typing import Callable

import copy
import numpy as np


class Trace(Sample):
    """Trace (or waveform) data from a single LCAx package"""

    def __init__(self, index, datasource, virtual_raw=None, virtual_ts=None):
        super().__init__(index, datasource, virtual_raw, virtual_ts)

    @property
    def raw(self):
        if self._raw is None:
            self._raw = super().raw

            for attribute in ['time_base_delays', 'distance_scaling', 'trace_smoothing_kernel']:
                if attribute not in self._raw and hasattr(self.datasource.sensor, attribute):
                    self._raw[attribute] = getattr(
                        self.datasource.sensor, attribute)

        return self._raw

    @property
    def raw_array(self):
        specs = self.specs
        raw = self.raw['data']
        return raw.reshape(specs['v'], specs['h'], raw.shape[-1])

    def processed_array(self, trace_processing: Callable):
        specs = self.specs
        processed = self.processed(trace_processing)['data']
        return processed.reshape(specs['v'], specs['h'], processed.shape[-1])

    @property
    def specs(self):
        # override the sensor specs if they are present in the YAML config file
        sensor_specs = self.datasource.sensor.specs
        if sensor_specs is not None:
            return sensor_specs
        return {k: self.raw[k] for k in ['v', 'h', 'v_fov', 'h_fov']}

    @property
    def timestamps(self):
        return self.raw['t']

    def processed(self, trace_processing: Callable):
        raw_copy = copy.deepcopy(self.raw)
        processed_traces = trace_processing(raw_copy)
        return processed_traces

    @staticmethod
    def saturation_flags(traces):
        flags = np.zeros(traces['data'].shape[0], dtype='u2')

        traces['data'] = traces['data'].astype('float64')
        saturation_value = traces['data'].max()
        if saturation_value == 0:
            return traces

        where_plateau = np.where(traces['data'] == saturation_value)
        channels, ind, sizes = np.unique(
            where_plateau[0], return_index=True, return_counts=True)
        positions = where_plateau[1][ind]

        for channel, position, size in zip(channels, positions, sizes):
            if size > 5 and position > 2 and position + size + 2 < traces['data'].shape[-1]:
                flags[channel] = 1
        return flags

    @property
    def max_range(self):
        raw = self.raw
        trace_lenght = raw['data'].shape[-1]
        return raw['time_base_delays'] + trace_lenght*raw['distance_scaling']

    @property
    def signal_to_noise(self):
        traces_zeroed = self.raw['data'] - \
            np.mean(self.raw['data'], axis=1)[:, None]
        return np.log10(np.mean(traces_zeroed**2, axis=1)/np.std(traces_zeroed, axis=1))+1



from pioneer.common import platform
from pioneer.common.trace_processing import Binning, Decimate, TraceProcessingCollection, ZeroBaseline
from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource
from pioneer.das.api.datatypes import datasource_xyzit_float_intensity
from pioneer.das.api.samples import FastTrace, PointCloud

from typing import Any

import numpy as np


class WaveformCloud(VirtualDatasource):
    """Makes a point cloud from each data point of each waveforms."""

    def __init__(self, reference_sensor:str, dependencies:list, zero_baseline:bool=False, remove_negative_distances:bool=True, decimation:int=1, binning:int=1):
        """Constructor
            Args:
                reference_sensor (str): The name of the sensor (e.g. 'pixell_bfc').
                dependencies (list): A list of the datasource names. 
                    The only element should be a Trace datasource (e.g. 'pixell_bfc_ftrr')
                zero_baseline (bool): If True, the baseline of the waveforms are re-calibrated to be around zero.
                remove_negative_distances (bool): If True, the points behind the sensor are filtered out.
                decimation (int): Must be an integer larger than 1. Downsample by this factor by decimation.
                binning (int): Must be an integer larger than 1. Downsample by this factor by binning.
        """
        trr_ds_name = dependencies[0].split('_')[-1]
        super(WaveformCloud, self).__init__(f'xyzit-{trr_ds_name}', dependencies, None)
        self.reference_sensor = reference_sensor
        self.original_trace_datasource = dependencies[0]
        self.remove_negative_distances = remove_negative_distances

        trace_processing_list = []
        if zero_baseline:
            trace_processing_list.append(ZeroBaseline())
        if binning != 1:
            trace_processing_list.append(Binning(binning))
        if decimation != 1:
            trace_processing_list.append(Decimate(decimation))
        self.trace_processing = TraceProcessingCollection(trace_processing_list)       

    def get_extrema_positions(self, sensor, distances, ts, time_base_delays):

        indices = np.arange(sensor.specs['v']*sensor.specs['h'])

        xyz0 = sensor.get_corrected_cloud(ts, sensor.cache(sensor.specs), 'point_cloud', indices, time_base_delays)
        x0, y0, z0 = xyz0[:,0], xyz0[:,1], xyz0[:,2]

        xyz1 = sensor.get_corrected_cloud(ts, sensor.cache(sensor.specs), 'point_cloud', indices, distances)
        x1, y1, z1 = xyz1[:,0], xyz1[:,1], xyz1[:,2]

        return x0, y0, z0, x1, y1, z1

    def __getitem__(self, key:Any):

        if isinstance(key, slice):
            return self[platform.slice_to_range(key, len(self))]
        if isinstance(key, range):
            return [self[index] for index in key]

        trace_sample = self.datasources[self.original_trace_datasource][key]
        timestamp = trace_sample.timestamp

        if isinstance(trace_sample, FastTrace):
            raw = trace_sample.processed(self.trace_processing)['high']
        else:
            raw = trace_sample.processed(self.trace_processing)

        traces = raw['data']

        x0, y0, z0, x1, y1, z1 = self.get_extrema_positions(
            sensor=self.datasources[self.original_trace_datasource].sensor, 
            distances=trace_sample.max_range, 
            ts=timestamp, 
            time_base_delays=raw['time_base_delays']
        )

        grid = np.indices(traces.shape)
        positions = grid[1].flatten()
        indices = grid[0].flatten()
        amplitudes = traces.flatten()
        timestamps = raw['timestamps'][indices]

        x = x0[indices]*(traces.shape[-1]-positions)/traces.shape[-1] + x1[indices]*positions/traces.shape[-1]
        y = y0[indices]*(traces.shape[-1]-positions)/traces.shape[-1] + y1[indices]*positions/traces.shape[-1]
        z = z0[indices]*(traces.shape[-1]-positions)/traces.shape[-1] + z1[indices]*positions/traces.shape[-1]

        distances = (x**2 + y**2 + z**2)**0.5

        if self.remove_negative_distances:
            keep = np.where(positions > -raw['time_base_delays'][indices]/raw['distance_scaling'])[0]
        else:
            keep = slice(None)

        virtual_raw = np.empty(indices[keep].size, dtype=datasource_xyzit_float_intensity())
        virtual_raw['x'] = x[keep]
        virtual_raw['y'] = y[keep]
        virtual_raw['z'] = z[keep]
        virtual_raw['i'] = amplitudes[keep]
        virtual_raw['t'] = timestamps[keep]

        return PointCloud(key, self, virtual_raw, timestamp)


from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource
from pioneer.das.api.samples import Sample

from typing import Any, List, Optional

import numpy as np


class BoundingBoxMetadata(VirtualDatasource):
    """Extract metadata from bounding boxes"""

    def __init__(self, reference_sensor:str, dependencies:List[str], category_counts:Optional[List[str]]=None):
        """Constructor
            Args:
                reference_sensor (str): The name of the reference sensor (e.g. 'pixell_bfc').
                dependencies (list): A list of the bounding boxes datasources. 
        """
        ds_type = 'scalars'
        super(BoundingBoxMetadata, self).__init__(ds_type, dependencies, None)
        self.reference_sensor = reference_sensor  
        self.reference_datasource = dependencies[0]
        self.category_counts = category_counts

        dtype = []
        
        if self.category_counts:
            dtype += [(category_name, 'u8') for category_name in self.category_counts]

        self.dtype = np.dtype(dtype)

    def get_at_timestamp(self, timestamp):
        sample = self.datasources[self.reference_datasource].get_at_timestamp(timestamp)
        return self[int(np.round(sample.index))]

    def __getitem__(self, key:Any):

        try: #is key iterable?
            return [self[index] for index in key]
        except: pass

        raw = np.zeros(1, self.dtype)

        for ds_name in self.dependencies:
            sample = self.datasources[ds_name][key]

            for category_name in sample.get_categories():
                if category_name in self.category_counts: raw[category_name] += 1

        return Sample(index=key, datasource=self, virtual_raw=raw, virtual_ts=sample.timestamp)




from pioneer.das.api.samples.point_cloud import PointCloud

from typing import Optional, Tuple

import numpy as np


class XYZV(PointCloud):
    """Point cloud provided by a radar"""

    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super().__init__(index, datasource, virtual_raw, virtual_ts)

    @property
    def fields(self) -> Tuple[str]:
        return self.raw[0].dtype.names

    def get_field(self, field: str) -> Optional[np.ndarray]:
        if field in self.fields: 
            return self.raw[0][field]

    @property
    def size(self) -> int:
        return self.raw[0].shape[0]


from pioneer.common import clouds, peak_detector
from pioneer.common.trace_processing import Desaturate, RemoveStaticNoise, Smooth, TraceProcessingCollection, ZeroBaseline
from pioneer.das.api.datasources.virtual_datasources.virtual_datasource import VirtualDatasource
from pioneer.das.api.samples import Echo, FastTrace

from typing import Any

import copy
import numpy as np



class Echoes_from_Traces(VirtualDatasource):
    """Non official implementation of the peak detector. Also measures the widths and the skews for each echo."""

    def __init__(self, reference_sensor:str, dependencies:list, nb_detections_max:int=3, min_amplitude:float=0, amplitude_factor:float=1.0):
        """Constructor
            Args:
                reference_sensor (str): The name of the sensor (e.g. 'pixell_bfc').
                dependencies (list): A list of the datasource names. 
                    The only element should be a Trace datasource (e.g. 'pixell_bfc_ftrr') 
                nb_detections_max (int): The maximum number of echoes per waveform.
                min_amplitude (float): Amplitude threshold under which the echoes are filtered out.
                amplitude_factor (float): Multiply amplitudes by a constant. Default is 1.
        """
        trr_ds_name = dependencies[0].split('_')[-1].split('-')[-1]
        super(Echoes_from_Traces, self).__init__(f'ech-{trr_ds_name}', dependencies, None)
        self.reference_sensor = reference_sensor
        self.original_trace_datasource = dependencies[0]
        self.nb_detections_max = nb_detections_max
        self.peak_detector = peak_detector.PeakDetector(nb_detections_max=self.nb_detections_max, min_amplitude=min_amplitude)
        self.amplitude_factor = amplitude_factor

        self.trace_processing = None

    def _set_trace_processing(self):
        sensor = self.datasources[self.dependencies[0]].sensor
        self.trace_processing = TraceProcessingCollection([
            Desaturate(sensor.saturation_calibration),
            RemoveStaticNoise(sensor.static_noise),
            ZeroBaseline(),
            Smooth(),
        ])

    def initialize_local_cache(self, data):
        self.local_cache = copy.deepcopy(data)

    def get_echoes(self, processed_traces):
        try:
            data = processed_traces
            self.local_cache['data'][...] = data['data']
            self.local_cache['time_base_delays'][...] = data['time_base_delays']
            self.local_cache['distance_scaling'][...] = data['distance_scaling']
        except:
            self.initialize_local_cache(processed_traces)

        traces = {'data': self.local_cache['data'], 
                  'time_base_delays': self.local_cache['time_base_delays'],
                  'distance_scaling': self.local_cache['distance_scaling']}

        echoes = self.peak_detector(traces)
        echoes['amplitudes'] *= self.amplitude_factor

        additionnal_fields = {}
        for key in echoes:
            if key not in ['indices','distances','amplitudes','timestamps','flags']:
                additionnal_fields[key] = [echoes[key], 'f4']
        return echoes, additionnal_fields

    def get_echoes_from_fast_traces(self, processed_fast_traces):
        # TODO: improve merging by replacing the saturated lines and columns
        sensor = self.datasources[self.dependencies[0]].sensor
        echoes_high, additionnal_fields_high = self.get_echoes(processed_fast_traces['high'])
        # echoes_low, additionnal_fields_low = self.get_echoes(processed_fast_traces['low'])
        echoes = {}
        for field in ['indices','distances','amplitudes']:
            # echoes[field] = np.hstack([echoes_high[field], echoes_low[field]])
            echoes[field] = echoes_high[field]
        additionnal_fields = {}
        for field in additionnal_fields_high:
            if field not in ['indices','distances','amplitudes']:
                # additionnal_fields[field] = [np.hstack([additionnal_fields_high[field][0], additionnal_fields_low[field][0]])
                #                             , additionnal_fields_high[field][1]]
                additionnal_fields[field] = additionnal_fields_high[field]
        return echoes, additionnal_fields

    def get_at_timestamp(self, timestamp):
        sample = self.datasources[self.original_trace_datasource].get_at_timestamp(timestamp)
        return self[int(np.round(sample.index))]

    def __getitem__(self, key:Any):

        if self.trace_processing is None:
            self._set_trace_processing()

        #Load data in local cache to prevent modifying the original data
        trace_sample = self.datasources[self.original_trace_datasource][key]
        timestamp = trace_sample.timestamp
        specs = trace_sample.specs
        if isinstance(trace_sample, FastTrace):
            get_echoes = self.get_echoes_from_fast_traces
        else:
            get_echoes = self.get_echoes

        processed_traces = trace_sample.processed(self.trace_processing)
        echoes, additionnal_fields = get_echoes(processed_traces)

        raw = clouds.to_echo_package(
            indices = np.array(echoes['indices'], 'u4'), 
            distances = np.array(echoes['distances'], 'f4'), 
            amplitudes = np.array(echoes['amplitudes'], 'f4'),
            additionnal_fields = additionnal_fields,
            timestamps = np.full(echoes['indices'].shape, timestamp), 
            flags = None, 
            timestamp = timestamp,
            specs = {"v" : specs['v'], "h" : specs['h'], "v_fov" : specs['v_fov'], "h_fov" : specs['h_fov']},
            distance_scale = 1.0, 
            amplitude_scale = 1.0, 
            led_power = 1.0, 
            eof_timestamp = None)
        
        return Echo(trace_sample.index, self, virtual_raw = raw, virtual_ts = timestamp)




from pioneer.das.api.samples.sample import Sample

from typing import Tuple

import cv2
import numpy as np
import warnings
warnings.simplefilter('once', DeprecationWarning)


class Image(Sample):
    """RGB data from a single camera image"""

    def __init__(self, index, datasource, virtual_raw = None, virtual_ts = None):
        super().__init__(index, datasource, virtual_raw, virtual_ts)
        self._und_camera_matrix = None

    def get_image(self, undistort:bool=False) -> np.ndarray:
        if type(self.raw) == dict: image = self.raw['image']
        image = self.raw
        if undistort:
            image = cv2.undistort(image, 
                cameraMatrix=self.camera_matrix, 
                distCoeffs=self.distortion_coeffs, 
                newCameraMatrix=self.und_camera_matrix,
            )
        return image

    @property
    def shape(self) -> Tuple[int, int, int]:
        return self.get_image().shape

    @property
    def camera_matrix(self):
        matrix = self.datasource.sensor.camera_matrix
        return matrix

    @property
    def distortion_coeffs(self):
        k = self.datasource.sensor.distortion_coeffs
        return k
    
    @property
    def und_camera_matrix(self):
        ''' the undistorted new camera matrix'''
        if self._und_camera_matrix is not None:
            return self._und_camera_matrix
        
        v,h,_ = self.shape
        self._und_camera_matrix, _ = cv2.getOptimalNewCameraMatrix(
            self.camera_matrix, self.distortion_coeffs, (h,v), 0.0, (h,v))
        return self._und_camera_matrix

    def project_pts(self, pts, mask_fov=False, output_mask=False, undistorted=False, margin=0):

        '''projects 3D points from camera referential to the image plane.

            Args:
                pts - (M,3)
                mask_fov - (True/False) If True the points outside fov are masked
                output_mask - (True/False) If True, the mask from mask_fov is returned
                undistorted - (True/False) If the image is undistorted
                margin (optionnal): margin (in pixels) outside the image unaffected by the fov mask
        '''

        R = T = np.zeros((3, 1))
        
        if undistorted or mask_fov or output_mask:
            A, k = self.und_camera_matrix, np.zeros((5,1))
            und_image_pts, _ = cv2.projectPoints(pts, R, T, A, k)
            und_image_pts = np.squeeze(und_image_pts)
            
        if not undistorted:
            A, k = self.camera_matrix, self.distortion_coeffs
            image_pts, _ = cv2.projectPoints(pts, R, T, A, k)
            image_pts = np.squeeze(image_pts)
        else:
            image_pts = und_image_pts

        if mask_fov or output_mask:
            mask = self.projection_mask(pts, und_image_pts, margin)

        if mask_fov:
            mask &= self.projection_mask(pts, image_pts, margin)
            image_pts = image_pts[mask]
            
        if output_mask:
            return image_pts, mask
        return image_pts
        
    def raw_image(self):
        warnings.warn("Image.raw_image() is deprecated. Use Image.get_image() instead.", DeprecationWarning)
        return self.get_image()
    
    def undistort_image(self):
        warnings.warn("Image.undistort_image() is deprecated. Use Image.get_image(undistort=True) instead.", DeprecationWarning)
        return self.get_image(undistort=True)

    def projection_mask(self, pts, projection, margin=0):
        v,h,_ = self.shape
        mask = (pts[:,2] > 0)& \
            (projection[:,0] >= 0 - margin) & \
            (projection[:,0] < h + margin) & \
            (projection[:,1] >= 0 - margin) & \
            (projection[:,1] < v + margin)
        return mask
                                                    



from pioneer.common.logging_manager import LoggingManager
from pioneer.das.api.samples.trace import Trace

from typing import Callable

import copy
import numpy as np


class FastTrace(Trace):
    """Derivation of Trace sample, for Pixell sensors. 
        Two distinct sets of waveforms are contained in the raw data dictionnary, under the keys 'high' and 'low'.
    """

    def __init__(self, index, datasource, virtual_raw=None, virtual_ts=None):
        super().__init__(index, datasource, virtual_raw, virtual_ts)

    @property
    def raw(self):
        if self._raw is None:
            self._raw = super(Trace, self).raw

            for fast_trace_type in ['low', 'high']:
                if 'time_base_delays' not in self._raw[fast_trace_type]:
                    self._raw[fast_trace_type]['time_base_delays'] = self.datasource.sensor.time_base_delays[fast_trace_type]
                self._raw[fast_trace_type]['distance_scaling'] = self.datasource.sensor.distance_scaling
                self._raw[fast_trace_type]['trace_smoothing_kernel'] = self.datasource.sensor.get_trace_smoothing_kernel()[
                    fast_trace_type]

        return self._raw

    @property
    def raw_array(self):
        specs = self.specs
        rawLow = self.raw['low']['data']
        rawMid = self.raw['high']['data']
        array = np.zeros((2, specs['v']*specs['h'], rawMid.shape[-1]))
        array[0] = rawMid
        array[1, :, :rawLow.shape[-1]] = rawLow
        return array.reshape((2, specs['v'], specs['h'], array.shape[-1]))

    def processed_array(self, trace_processing: Callable):
        specs = self.specs
        processedLow = self.processed(trace_processing)['low']['data']
        processedMid = self.processed(trace_processing)['high']['data']
        array = np.zeros((2, specs['v']*specs['h'], processedMid.shape[-1]))
        array[0] = processedMid
        array[1, :, :processedLow.shape[-1]] = processedLow
        return array.reshape((2, specs['v'], specs['h'], array.shape[-1]))

    def processed(self, trace_processing: Callable):
        processed_traces = {}
        raw_copy = copy.deepcopy(self.raw)
        for fast_trace_type in ['low', 'high']:
            processed_traces[fast_trace_type] = trace_processing(
                raw_copy[fast_trace_type])
        return processed_traces

    @property
    def max_range(self):
        raw = self.raw
        trace_lenght = raw['high']['data'].shape[-1]
        return raw['high']['time_base_delays'] + trace_lenght*raw['high']['distance_scaling']

    @property
    def signal_to_noise(self):
        snr = {}
        for fast_trace_type in ['low', 'high']:
            traces_zeroed = self.raw[fast_trace_type]['data'] - \
                np.mean(self.raw[fast_trace_type]['data'], axis=1)[:, None]
            snr[fast_trace_type] = np.log10(
                np.mean(traces_zeroed**2, axis=1)/np.std(traces_zeroed, axis=1))+1
        return snr
